/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Managment Committee (PMC)
 * (C) 2000, Institut de Recherche pour le Dveloppement
 * (C) 1999, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.axis;

// J2SE dependencies
import java.text.Format;
import java.util.Locale;
import java.awt.Font;
import java.awt.RenderingHints;
import java.beans.PropertyChangeListener;

// Geotools dependencies
import org.geotools.units.Unit;


/**
 * An axis's graduation. A <code>Graduation</code> object encompass minimal
 * and maximal values for an axis in arbitrary units, and allow access to
 * tick locations and labels through a {@link TickIterator} object.
 *
 * Different implementations may compute tick locations in different ways.
 * For example a graduation for dates is handled in a different way than a
 * graduation for numbers.
 *
 * @version $Id: Graduation.java,v 1.3 2003/05/13 10:58:46 desruisseaux Exp $
 * @author Martin Desruisseaux
 */
public interface Graduation {
    /**
     * Rendering hint for the axis length, in pixels or points (1/72 of inch).
     * Values for this key must be {@link Number} objects. This hint is used
     * together with {@link #VISUAL_TICK_SPACING} during {@link TickIterator}
     * creation in order to compute a tick increment value.
     *
     * @see #getTickIterator
     */
    RenderingHints.Key VISUAL_AXIS_LENGTH = new RenderingHintKey(Number.class, 0);

    /**
     * Rendering hint for the preferred spacing between ticks, in pixels or points
     * (1/72 of inch). Values for this key must be {@link Number} objects. This hint
     * is used together with {@link #VISUAL_AXIS_LENGTH} during {@link TickIterator}
     * creation in order to compute a tick increment value. The tick spacing really
     * used may be slightly different, since {@link TickIterator} may choose a rounded
     * value.
     *
     * @see #getTickIterator
     */
    RenderingHints.Key VISUAL_TICK_SPACING = new RenderingHintKey(Number.class, 1);

    /**
     * The font to use for rendering tick labels. Value for this key must be a {@link Font}
     * object. If this hint is not provided, a default font will be used.
     *
     * @see Axis2D#paint
     */
    RenderingHints.Key TICK_LABEL_FONT = new RenderingHintKey(Font.class, 2);

    /**
     * The font to use for rendering the title. Value for this key must be a {@link Font} object.
     * If this hint is not provided, a default font will be used.
     *
     * @see Axis2D#paint
     */
    RenderingHints.Key AXIS_TITLE_FONT = new RenderingHintKey(Font.class, 3);

    /**
     * Returns the minimal value for this graduation.
     * @return The minimal value in {@link #getUnit} units.
     *
     * @see #getMaximum
     * @see #getRange
     */
    public abstract double getMinimum();

    /**
     * Returns the maximal value for this graduation.
     * @return The maximal value in {@link #getUnit} units.
     *
     * @see #getMinimum
     * @see #getRange
     */
    public abstract double getMaximum();

    /**
     * Returns the graduation's range. This is equivalents to computing
     * <code>{@link #getMaximum}-{@link #getMinimum}</code>. However,
     * some implementation may optimize this computation in order to
     * avoid rounding errors.
     */
    public abstract double getRange();

    /**
     * Returns the axis title. If <code>includeUnits</code> is <code>true</code>,
     * then the returned string will includes units as in "Temperature (C)", or
     * time zone as in "Start time (UTC)". The exact formatting is local-dependent.
     *
     * @param  includeSymbol <code>true</code> to format the unit or timezone symbol after the name.
     * @return The graduation name (also to be use as axis title).
     */
    public abstract String getTitle(final boolean includeSymbol);

    /**
     * Returns the graduation's units, or <code>null</code> if unknow.
     */
    public abstract Unit getUnit();

    /**
     * Returns the locale to use for formatting title and labels.
     */
    public abstract Locale getLocale();

    /**
     * Returns the format to use for formatting labels. The format really used by
     * {@link TickIterator#currentLabel} may not be the same. For example, some
     * iterators may adjust automatically the number of fraction digits.
     */
    public abstract Format getFormat();

    /**
     * Returns an iterator object that iterates along the graduation ticks
     * and provides access to the graduation values. If an optional {@link
     * RenderingHints} is specified, tick locations are adjusted according
     * values for {@link #VISUAL_AXIS_LENGTH} and {@link #VISUAL_TICK_SPACING}
     * keys.
     *
     * @param  hints Rendering hints for the axis, or <code>null</code> for
     *         the default hints.
     * @param  reuse An iterator to reuse if possible, or <code>null</code>
     *         to create a new one. A non-null object may help to reduce the
     *         number of object garbage-collected when rendering the axis.
     * @return A iterator to use for iterating through the graduation. This
     *         iterator may or may not be the <code>reuse</code> object.
     */
    public abstract TickIterator getTickIterator(RenderingHints hints, TickIterator reuse);

    /**
     * Adds a {@link PropertyChangeListener} to the listener list.
     * The listener is registered for all properties, such as "label"
     * and "locale".
     */
    public abstract void addPropertyChangeListener(PropertyChangeListener listener);

    /**
     * Removes a {@link PropertyChangeListener} from the listener list.
     */
    public abstract void removePropertyChangeListener(PropertyChangeListener listener);
}
