/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Management Committee (PMC)
 * (C) 2001, Institut de Recherche pour le Dveloppement
 * (C) 1999, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.gui.headless;

// Gestion des entrs/sorties
import java.lang.System;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.UnsupportedCharsetException;

// Gestion du texte
import java.text.NumberFormat;
import java.text.BreakIterator;

// Divers
import org.geotools.util.ProgressListener;
import org.geotools.resources.Arguments;
import org.geotools.resources.Utilities;
import org.geotools.resources.gui.Resources;
import org.geotools.resources.gui.ResourceKeys;


/**
 * Informe l'utilisateur des progrs d'une opration  l'aide de messages envoy vers un
 * flot. L'avancement de l'opration sera affich en pourcentage sur une ligne (gnralement
 * le priphrique de sortie standard). Cette classe peut aussi crire des avertissements,
 * ce qui est utile entre autre lors de la lecture d'un fichier de donnes durant laquelle
 * on veut signaler des anomalies mais sans arrter la lecture pour autant.
 *
 * @version $Id: ProgressPrinter.java,v 1.2 2003/05/13 11:01:39 desruisseaux Exp $
 * @author Martin Desruisseaux
 */
public class ProgressPrinter implements ProgressListener {
    /**
     * Nom de l'opration en cours. Le pourcentage sera cris  la droite de ce nom.
     */
    private String description;

    /**
     * Flot utilis pour l'criture de l'tat d'avancement d'un
     * processus ainsi que pour les critures des commentaires.
     */
    private final PrintWriter out;

    /**
     * Indique si le caractre '\r' ramne au dbut de la ligne courante sur
     * ce systme. On supposera que ce sera le cas si le systme n'utilise
     * pas la paire "\r\n" pour changer de ligne (comme le system VAX-VMS).
     */
    private final boolean CR_supported;

    /**
     * Longueur maximale des lignes. L'espace utilisable sera un peu
     * moindre car quelques espaces seront laisss en dbut de ligne.
     */
    private final int maxLength;

    /**
     * Nombre de caractres utiliss lors de l'criture de la dernire ligne.
     * Ce champ est mis  jour par la mthode {@link #carriageReturn} chaque
     * fois que l'on dclare que l'on vient de terminer l'criture d'une ligne.
     */
    private int lastLength;

    /**
     * Position  laquelle commencer  crire le pourcentage. Cette information
     * est gre automatiquement par la mthode {@link #progress}. La valeur -1
     * signifie que ni le pourcentage ni la description n'ont encore t crits.
     */
    private int percentPosition = -1;

    /**
     * Dernier pourcentage crit. Cette information est utilise
     * afin d'viter d'crire deux fois le mme pourcentage, ce
     * qui ralentirait inutilement le systme. La valeur -1 signifie
     * qu'on n'a pas encore crit de pourcentage.
     */
    private float lastPercent = -1;

    /**
     * Format  utiliser pour crire les pourcentages.
     */
    private NumberFormat format;

    /**
     * Objet utilis pour couper les lignes correctements lors de l'affichage
     * de messages d'erreurs qui peuvent prendre plusieurs lignes.
     */
    private BreakIterator breaker;

    /**
     * Indique si cet objet a dj crit des avertissements. Si
     * oui, on ne rcrira pas le gros titre "avertissements".
     */
    private boolean hasPrintedWarning;

    /**
     * Source du dernier message d'avertissement. Cette information est
     * conserve afin d'viter de rpter la source lors d'ventuels
     * autres messages d'avertissements.
     */
    private String lastSource;

    /**
     * Construit un objet qui crira sur le priphrique de sortie standard
     * ({@link java.lang.System#out}) l'tat d'avancement d'une opration.
     * La longueur par dfaut des lignes sera de 80 caractres.
     */
    public ProgressPrinter() {
        this(new PrintWriter(Arguments.getWriter(System.out)));
    }

    /**
     * Construit un objet qui crira sur le priphrique de
     * sortie spcifi l'tat d'avancement d'une opration.
     * La longueur par dfaut des lignes sera de 80 caractres.
     */
    public ProgressPrinter(final PrintWriter out) {
        this(out, 80);
    }

    /**
     * Construit un objet qui crira sur le priphrique de
     * sortie spcifi l'tat d'avancement d'une opration.
     *
     * @param out priphrique de sortie  utiliser pour crire l'tat d'avancement.
     * @param maxLength Longueur maximale des lignes. Cette information est utilise
     *        par {@link #warningOccurred} pour rpartir sur plusieurs lignes des
     *        messages qui ferait plus que la longueur <code>lineLength</code>.
     */
    public ProgressPrinter(final PrintWriter out, final int maxLength) {
        this.out = out;
        this.maxLength = maxLength;
        final String lineSeparator = System.getProperty("line.separator");
        CR_supported=(lineSeparator!=null && lineSeparator.equals("\r\n"));
    }

    /**
     * Efface le reste de la ligne (si ncessaire) puis repositionne le curseur au dbut
     * de la ligne. Si les retours chariot ne sont pas supports, alors cette mthode va
     * plutt passer  la ligne suivante. Dans tous les cas, le curseur se trouvera au
     * dbut d'une ligne et la valeur <code>length</code> sera affect au champ
     * {@link #lastLength}.
     *
     * @param length Nombre de caractres qui ont t crit jusqu' maintenant sur cette ligne.
     *        Cette information est utilise pour ne mettre que le nombre d'espaces ncessaires
     *         la fin de la ligne.
     */
    private void carriageReturn(final int length) {
        if (CR_supported && length<maxLength) {
            for (int i=length; i<lastLength; i++)  {
                out.print(' ');
            }
            out.print('\r');
            out.flush();
        } else {
            out.println();
        }
        lastLength = length;
    }

    /**
     * Ajoute des points  la fin de la ligne jusqu' reprsenter
     * le pourcentage spcifi. Cette mthode est utilise pour
     * reprsenter les progrs sur un terminal qui ne supporte
     * pas les retours chariots.
     *
     * @param percent Pourcentage accompli de l'opration. Cette
     *        valeur doit obligatoirement se trouver entre 0 et
     *        100 (a ne sera pas vrifi).
     */
    private void completeBar(final float percent) {
        final int end = (int) ((percent/100)*((maxLength-2)-percentPosition)); // Round toward 0.
        while (lastLength < end) {
            out.print('.');
            lastLength++;
        }
    }

    /**
     * Retourne le message d'crivant l'opration
     * en cours. Si aucun message n'a t dfinie,
     * retourne <code>null</code>.
     */
    public String getDescription() {
        return description;
    }

    /**
     * Spcifie un message qui dcrit l'opration en cours.
     * Ce message est typiquement spcifie avant le dbut
     * de l'opration. Toutefois, cette mthode peut aussi
     * tre appele  tout moment pendant l'opration sans
     * que cela affecte le pourcentage accompli. La valeur
     * <code>null</code> signifie qu'on ne souhaite plus
     * afficher de description.
     */
    public synchronized void setDescription(final String description) {
        this.description = description;
    }

    /**
     * Indique que l'opration a commence.
     */
    public synchronized void started() {
        int length = 0;
        if (description != null) {
            out.print(description);
            length=description.length();
        }
        if (CR_supported) {
            carriageReturn(length);
        }
        out.flush();
        percentPosition   = length;
        lastPercent       = -1;
        lastSource        = null;
        hasPrintedWarning = false;
    }

    /**
     * Indique l'tat d'avancement de l'opration. Le progrs est reprsent par un
     * pourcentage variant de 0  100 inclusivement. Si la valeur spcifie est en
     * dehors de ces limites, elle sera automatiquement ramene entre 0 et 100.
     */
    public synchronized void progress(float percent) {
        if (percent<0  ) percent=0;
        if (percent>100) percent=100;
        if (CR_supported) {
            /*
             * Si le priphrique de sortie supporte les retours chariot,
             * on crira l'tat d'avancement comme un pourcentage aprs
             * la description, comme dans "Lecture des donnes (38%)".
             */
            if (percent != lastPercent) {
                if (format == null) {
                    format = NumberFormat.getPercentInstance();
                }
                final String text = format.format(percent/100.0);
                int length = text.length();
                percentPosition = 0;
                if (description != null) {
                    out.print(description);
                    out.print(' ');
                    length += (percentPosition=description.length())+1;
                }
                out.print('(');
                out.print(text);
                out.print(')');
                length += 2;
                carriageReturn(length);
                lastPercent=percent;
            }
        } else {
            /*
             * Si le priphrique ne supporte par les retours chariots, on
             * crira l'tat d'avancement comme une srie de points placs
             * aprs la description, comme dans "Lecture des donnes......"
             */
            completeBar(percent);
            lastPercent=percent;
            out.flush();
        }
    }

    /**
     * Indique que l'opration est termine. L'indicateur visuel informant des
     * progrs sera ramen  100% ou disparatra. Si des messages d'erreurs ou
     * d'avertissements taient en attente, ils seront crits.
     */
    public synchronized void complete() {
        if (!CR_supported) {
            completeBar(100);
        }
        carriageReturn(0);
        out.flush();
    }

    /**
     * Libre les ressources utilises par cet objet.
     * L'implmentation par dfaut ne fait rien.
     */
    public void dispose() {
    }

    /**
     * Envoie un message d'avertissement. La premire fois que cette mthode est appelle, le mot
     * "AVERTISSEMENTS" sera crit en lettres majuscules au milieu d'une bote. Si une source est
     * spcifie (argument <code>source</code>), elle ne sera crite qu' la condition qu'elle
     * n'est pas la mme que celle du dernier avertissement. Si une note de marge est spcifie
     * (argument <code>margin</code>), elle sera crite entre parenthses  la gauche de
     * l'avertissement <code>warning</code>.
     *
     * @param source Chane de caractre dcrivant la source de l'avertissement.
     *        Il s'agira par exemple du nom du fichier dans lequel une anomalie
     *        a t dtecte. Peut tre nul si la source n'est pas connue.
     * @param margin Texte  placer dans la marge de l'avertissement <code>warning</code>,
     *        ou <code>null</code> s'il n'y en a pas. Il s'agira le plus souvent du numro
     *        de ligne o s'est produite l'erreur dans le fichier <code>source</code>.
     * @param warning Message d'avertissement  crire. Si ce message est
     *        plus long que la largeur de l'cran (telle que spcifie au
     *        moment de la construction, alors il sera automatiquement
     *        distribu sur plusieurs lignes correctements indentes.
     */
    public synchronized void warningOccurred(final String source, String margin,
                                             final String warning)
    {
        carriageReturn(0);
        if (!hasPrintedWarning) {
            printInBox(Resources.format(ResourceKeys.WARNING));
            hasPrintedWarning=true;
        }
        if (!Utilities.equals(source, lastSource)) {
            out.println();
            out.println(source!=null ? source : Resources.format(ResourceKeys.UNTITLED));
            lastSource=source;
        }
        /*
         * Procde  l'criture de l'avertissement avec (de faon optionnelle)
         * quelque chose dans la marge (le plus souvent un numro de ligne).
         */
        String prefix="    ";
        String second=prefix;
        if (margin != null) {
            margin = trim(margin);
            if (margin.length() != 0) {
                final StringBuffer buffer = new StringBuffer(prefix);
                buffer.append('(');
                buffer.append(margin);
                buffer.append(") ");
                prefix=buffer.toString();
                buffer.setLength(0);
                second=Utilities.spaces(prefix.length());
            }
        }
        int width=maxLength-prefix.length()-1;
        if (breaker == null) {
            breaker=BreakIterator.getLineInstance();
        }
        breaker.setText(warning);
        int start=breaker.first(), end=start, nextEnd;
        while ((nextEnd=breaker.next()) != BreakIterator.DONE) {
            while (nextEnd-start > width) {
                if (end <= start) {
                    end=Math.min(nextEnd, start+width);
                }
                out.print(prefix);
                out.println(warning.substring(start, end));
                prefix=second;
                start=end;
            }
            end=Math.min(nextEnd, start+width);
        }
        if (end>start) {
            out.print(prefix);
            out.println(warning.substring(start, end));
        }
        if (!CR_supported && description!=null) {
            out.print(description);
            completeBar(lastPercent);
        }
        out.flush();
    }

    /**
     * Indique qu'une exception est survenue pendant le traitement de l'opration.
     * L'implmentation par dfaut crit "Exception" dans une bote, puis envoie
     * la trace vers le priphrique de sortie spcifie au constructeur.
     */
    public synchronized void exceptionOccurred(final Throwable exception) {
        carriageReturn(0);
        printInBox(Resources.format(ResourceKeys.EXCEPTION));
        exception.printStackTrace(out);
        hasPrintedWarning = false;
        out.flush();
    }

    /**
     * Retourne la chane <code>margin</code> sans les
     * ventuelles parenthses qu'elle pourrait avoir
     * de part et d'autre.
     */
    private static String trim(String margin) {
        margin = margin.trim();
        int lower = 0;
        int upper = margin.length();
        while (lower<upper && margin.charAt(lower+0)=='(') lower++;
        while (lower<upper && margin.charAt(upper-1)==')') upper--;
        return margin.substring(lower, upper);
    }

    /**
     * crit dans une bote entour d'astrix le texte spcifi en argument.
     * Ce texte doit tre sur une seule ligne et ne pas comporter de retour
     * chariot. Les dimensions de la bote seront automatiquement ajustes.
     * @param text Texte  crire (une seule ligne).
     */
    private void printInBox(String text) {
        int length = text.length();
        for (int pass=-2; pass<=2; pass++) {
            switch (Math.abs(pass)) {
                case 2: for (int j=-10; j<length; j++) out.print('*');
                        out.println();
                        break;

                case 1: out.print("**");
                        for (int j=-6; j<length; j++) out.print(' ');
                        out.println("**");
                        break;

                case 0: out.print("**   ");
                        out.print(text);
                        out.println("   **");
                        break;
            }
        }
    }
}
