/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Managment Committee (PMC)
 * (C) 2001, Institut de Recherche pour le Dveloppement
 * (C) 1999, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.renderer.array;

// Divers
import java.util.Iterator;
import java.awt.geom.Point2D;

// Geotools dependencies
import org.geotools.util.Cloneable;


/**
 * Itrateur balayant les donnes d'un tableau {@link PointArray}. Cet itrateur est obtenu par
 * un appel  {@link PointArray#iterator}.  Le balayage se fait gnralement par des appels aux
 * aux mthodes {@link #nextX} et {@link #nextY}, qui retournent les donnes sous forme de nombres
 * rels <code>float</code>. Toutefois, ces mthodes <g>doivent toujours</g> tre appelles dans
 * cet ordre: {@link #nextX} d'abord, suivit de {@link #nextY}. Tout manquement  cette rgle (par
 * exemple appeller {@link #nextX} deux fois de suite) peut produire des rsultats errones.
 *
 * @version $Id: PointIterator.java,v 1.5 2003/08/28 15:42:13 desruisseaux Exp $
 * @author Martin Desruisseaux
 *
 * @task TODO: Once J2SE 1.5 will be available, this class should implements
 *             Iterator<Point2D> and method next() should returns Point2D.
 *             Method clone() should returns PointIterator.
 */
public abstract class PointIterator implements Iterator, Cloneable {
    /**
     * Constructeur par dfaut.
     */
    protected PointIterator() {
    }

    /**
     * Indique si les mthodes {@link #next} peuvent retourner d'autres donnes.
     */
    public abstract boolean hasNext();

    /**
     * Retourne la valeur de la longitude courante. Avant d'appeller
     * une seconde fois cette mthode, il faudra <g>obligatoirement</g>
     * avoir appel {@link #nextY}.
     */
    public abstract float nextX();

    /**
     * Retourne la valeur de la latitude courante, puis avance au point
     * suivant. Chaque appel de cette mthode doit <g>obligatoirement</g>
     * avoir t prcde d'un appel  la mthode {@link #nextX}.
     */
    public abstract float nextY();

    /**
     * Retourne la valeur du point courant dans un objet {@link Point2D},
     * puis avance au point suivant. Cette mthode combine un appel de
     * {@link #nextX} suivit de {@link #nextY}.
     */
    public Object next() {
        return new Point2D.Float(nextX(), nextY());
    }

    /**
     * Opration non-supporte.
     */
    public final void remove() {
        throw new UnsupportedOperationException();
    }

    /**
     * Retourne une copie conforme de cet itrateur. Cette
     * copie peut tre utile pour balayer une seconde fois
     * les mmes donnes  partir du point o se trouve cet
     * itrateur.
     */
    public final Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException exception) {
            // Should not happen, since we are cloneable
            throw new AssertionError(exception);
        }
    }
}
