/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Managment Committee (PMC)
 * (C) 2001, Institut de Recherche pour le Dveloppement
 * (C) 1999, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.renderer.array;

// J2SE dependencies
import java.awt.geom.Point2D;


/**
 * Classe enveloppant une portion seulement d'un tableau <code>float[]</code>.
 * Des instances de cette classes sont retournes par {@link DefaultArray#subarray}.
 * L'implmentation par dfaut de cette classe est imutable. Toutefois, certaines
 * classes drives (notamment {@link DynamicArray}) ne le seront pas forcment.
 *
 * @version $Id: SubArray.java,v 1.5 2003/05/23 17:58:59 desruisseaux Exp $
 * @author Martin Desruisseaux
 */
class SubArray extends DefaultArray {
    /**
     * Numro de srie (pour compatibilit avec des versions antrieures).
     */
    private static final long serialVersionUID = 5172936367826790633L;

    /**
     * Plage des donnes valides du tableau {@link #array}.
     */
    protected int lower, upper;

    /**
     * Enveloppe une partie d'un tableau de <code>float[]</code>.
     *
     * @param  array Tableau de coordonnes (<var>x</var>,<var>y</var>).
     * @param  lower Index de la premire coordonnes <var>x</var> 
     *         prendre en compte dans le tableau <code>array</code>.
     * @param  upper Index suivant celui de la dernire coordonne <var>y</var> 
     *         prendre en compte dans le tableau <code>array</code>. La diffrence
     *         <code>upper-lower</code> doit obligatoirement tre paire.
     */
    public SubArray(final float[] array, final int lower, final int upper) {
        super(array);
        this.lower = lower;
        this.upper = upper;
        checkRange(array, lower, upper);
    }

    /**
     * Retourne l'index de la premire coordonne valide.
     */
    protected final int lower() {
        return lower;
    }

    /**
     * Retourne l'index suivant celui de la dernire coordonne valide.
     */
    protected final int upper() {
        return upper;
    }

    /**
     * Returns the point at the specified index.
     *
     * @param  index The index from 0 inclusive to {@link #count} exclusive.
     * @return The point at the given index.
     * @throws IndexOutOfBoundsException if <code>index</code> is out of bounds.
     */
    public final Point2D getValue(int index) throws IndexOutOfBoundsException {
        if (index >= 0) {
            index = 2*index + lower;
            if (index < upper) {
                return new Point2D.Float(array[index], array[index+1]);
            }
        }
        throw new IndexOutOfBoundsException();
    }

    /**
     * Returns an estimation of memory usage in bytes. This method returns the same value
     * than {@link DefaultArray#getMemoryUsage} plus 8 bytes for the internal fields (the
     * {@link #lower} and {@link #upper} fields).
     */
    public long getMemoryUsage() {
        return super.getMemoryUsage() + 8;
    }
}
