/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Managment Committee (PMC)
 * (C) 2003, Institut de Recherche pour le Dveloppement
 * (C) 1998, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.renderer.geom;


/**
 * Rfrence vers un objet {@link Polyline}. Cette classe contient aussi un valeur
 * boolenne qui sera prise en compte par la mthode {@link #hashCode}. Cette valeur boolenne
 * agit comme le spin d'un lectron. Deux instances de <code>Fermion</code> peuvent rfrer au
 * mme segment {@link Polyline} s'ils n'ont pas la mme valeur boolenne ("spin"). Cette classe
 * est rserve  un usage interne par {@link PolygonAssembler}.
 *
 * @see FermionPair
 *
 * @version $Id: Fermion.java,v 1.3 2003/05/27 18:22:43 desruisseaux Exp $
 * @author Martin Desruisseaux
 */
final class Fermion {
    /**
     * Rfrence vers la polyligne reprsente par cet objet. Dans l'analogie
     * avec la mcanique quantique, a serait le "niveau atomique" d'un Fermion.
     */
    Polyline path;

    /**
     * Si <code>true</code>, la fin de du trait de cte <code>path</code> devra tre fusionn
     * avec un autre trait (inconnu de cet objet). Si <code>false</code>, c'est le dbut du trait
     * <code>path</code> qui devra tre fusionn. Dans l'analogie avec la mcanique quantique,
     * c'est le "spin" d'un Fermion.
     */
    boolean mergeEnd;

    /**
     * Indique si deux cls sont identiques. Deux cls sont considrs identiques si elles
     * se rfrent au mme trait de cte {@link #path} avec la mme valeur boolenne
     * {@link #mergeEnd}.
     */
    public boolean equals(final Object o) {
        if (o instanceof Fermion) {
            final Fermion k=(Fermion) o;
            return k.path==path && k.mergeEnd==mergeEnd;
        } else {
            return false;
        }
    }

    /**
     * Retourne une valeur  peu prs unique pour cet objet. Cette valeur sera btie 
     * partir de la rfrence {@link #path} et de la valeur boolenne {@link #mergeEnd}.
     */
    public int hashCode() {
        final int code = System.identityHashCode(path);
        return mergeEnd ? code : ~code;
    }

    /**
     * Renvoie une reprsentation sous forme de chane de caractres de cet objet.
     * Cette reprsentation sera de la forme "Fermion[52 pts]".
     */
    public String toString() {
        final StringBuffer buffer = new StringBuffer("Fermion[");
        if (path != null) {
            buffer.append(path.getPointCount());
            buffer.append(" pts");
        }
        buffer.append(']');
        return buffer.toString();
    }
}
