/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Managment Committee (PMC)
 * (C) 2001, Institut de Recherche pour le Dveloppement
 * (C) 1998, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.renderer.geom;

// Collections
import java.util.List;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.AbstractCollection;
import java.util.NoSuchElementException;
import java.lang.UnsupportedOperationException;

// Geometry
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.GeneralPath;

// Arrays
import org.geotools.renderer.array.ArrayData;
import org.geotools.renderer.array.PointArray;
import org.geotools.renderer.array.DefaultArray;
import org.geotools.renderer.array.PointIterator;

// Coordinate systems
import org.geotools.units.Unit;
import org.geotools.cs.Ellipsoid;
import org.geotools.cs.CoordinateSystem;
import org.geotools.cs.CompoundCoordinateSystem;
import org.geotools.cs.ProjectedCoordinateSystem;
import org.geotools.cs.GeographicCoordinateSystem;
import org.geotools.cs.HorizontalDatum;
import org.geotools.ct.MathTransform;
import org.geotools.ct.MathTransform2D;
import org.geotools.ct.TransformException;
import org.geotools.ct.CoordinateTransformation;

// Miscellaneous
import java.io.Serializable;
import org.geotools.math.Statistics;
import org.geotools.resources.XArray;
import org.geotools.resources.Geometry;
import org.geotools.resources.Utilities;
import org.geotools.resources.CTSUtilities;
import org.geotools.resources.renderer.Resources;
import org.geotools.resources.renderer.ResourceKeys;


/**
 * Lignes traces sans lever le crayon. Ces lignes ne reprsentent par forcment une forme ferme
 * (un polygone). Les objets <code>LineString</code> ont deux caractristiques particulires:
 *
 * <ul>
 *   <li>Ils mmorisent sparment les points qui ne font que former une bordure. Par exemple, si
 *       seulement la moiti d'une le apparat sur une carte, les points qui servent  joindre
 *       les deux extrmits des polylignes (en suivant la bordure de la carte l o l'le est
 *       coupe) n'ont pas de ralit gographique. Dans chaque objet <code>LineString</code>, il doit
 *       y avoir une distinction claire entre les vritable points gographique les "points de
 *       bordure". Ces points sont mmoriss sparments dans les tableaux
 *       {@link #prefix}/{@link #suffix} et {@link #array} respectivement.</li>
 *
 *   <li>Ils peuvent tre chans avec d'autres objets <code>LineString</code>. Former une chane
 *       d'objets <code>LineString</code> peut tre utile lorsque les coordonnes d'une cte ont t
 *       obtenues  partir de la digitalisation de plusieurs cartes bathymtriques, que l'on joindra
 *       en une ligne continue au moment du traage. Elle peut aussi se produire lorsqu'une ligne
 *       qui se trouve prs du bord de la carte entre, sort, rentre et resort plusieurs fois du
 *       cadre.</li>
 * </ul>
 *
 * Par convention, toutes les mthodes statiques de cette classe peuvent agir
 * sur une chane d'objets {@link LineString} plutt que sur une seule instance.
 *
 * @version $Id: LineString.java,v 1.2 2003/07/23 14:17:32 desruisseaux Exp $
 * @author Martin Desruisseaux
 */
final class LineString implements Serializable {
    /**
     * Serial version for compatibility with previous version.
     */
    private static final long serialVersionUID = 719026440018802849L;

    /**
     * Set to <code>true</code> for disallowing two consecutive points with the same value.
     * Some algorithms in this class and in {@link PathAnalyser} require the distance between
     * two consecutive points to be greater than 0 in all cases.
     */
    private static final boolean REMOVE_DOUBLONS = true;

    /**
     * Set to <code>true</code> for removing doublons in borders as well.
     *
     * @task TODO: Current algorithm prevent the creation of curved line. The algorithm
     *             should be modified in such a way that  doublons are removed from the
     *             first of last points of the main data instead of from the border (the
     *             should not be modified by '[append|prepend]Border'. The work done by
     *             'addBorder' is correct however). How to do: add the following methods:
     *             'remove[First|Last]Points(int n)', and use it in '[append|prepend]Border'.
     */
    private static final boolean REMOVE_DOUBLONS_IN_BORDER = false;

    /**
     * Set to <code>true</code> if {@link #freeze} should try to merge the {@link #array} of
     * two consecutive polylines. Experience has show that it was often a bad idea, since it
     * force a lot of copies during clipping.
     */
    private static final boolean MERGE_POLYLINE_DATA = false;

    /**
     * Polylignes prcdentes et suivantes. La classe <code>LineString</code> implmente une liste 
     * double liens. Chaque objet <code>LineString</code> est capable d'accder et d'agir sur les
     * autres lments de la liste  laquelle il appartient. En consquent, il n'est pas ncessaire
     * d'utiliser une classe spare (par exemple {@link java.util.LinkedList}) comme conteneur.
     * Il ne s'agit pas forcment d'un bon concept de programmation, mais il est pratique dans le
     * cas particulier de la classe <code>LineString</code> et offre de bonnes performances.
     */
    private LineString previous, next;

    /**
     * Coordonnes formant la polyligne. Ces coordonnes doivent tre celles d'un trait de cte ou
     * de toute autre forme gomtrique ayant une signification cartographique. Les points qui
     * servent  "couper" un polygone (par exemple des points longeant la bordure de la carte)
     * doivent tre mmoriss sparment dans le tableau <code>suffix</code>.
     */
    private PointArray array;

    /**
     * Coordonnes  retourner aprs celles de <code>array</code>. Ces coordonnes servent
     * gnralement  refermer un polygone, par exemple en suivant le cadre de la carte. Ce
     * champ peut tre nul s'il ne s'applique pas.
     */
    private PointArray suffix;

    /**
     * Valeur minimales et maximales autorises comme arguments pour les mthodes {@link #getArray}
     * et {@link #setArray}. Lorsque ces valeurs sont utilises en ordre croissant, {@link #getArray}
     * retourne dans l'ordre les tableaux {@link #prefix}, {@link #array} et {@link #suffix}.
     * <br><br>
     * Note: si les valeurs de ces constantes changent, alors il faudra revoir l'implmentation des
     * mthodes suivantes:
     *
     *    {@link #getArray},
     *    {@link #setArray},
     *    {@link #reverse},
     *    {@link #freeze},
     */
    private static final int FIRST_ARRAY=0, LAST_ARRAY=1;

    /**
     * Construit un objet qui enveloppera les points spcifis.
     * Cette polyligne fera initialement partie d'aucune liste.
     */
    LineString(final PointArray array) {
        this.array = array;
    }

    /**
     * Construit des objets mmorisant les coordonnes <code>data</code>. Les valeurs
     * <code>NaN</code> au dbut et  la fin de <code>data</code> seront ignores. Celles
     * qui apparaissent au milieu auront pour effet de sparer le trait en plusieurs polylignes.
     *
     * @param data   Tableau de coordonnes (peut contenir des NaN).
     * @return       Tableau de polylignes. Peut avoir une longueur de 0, mais ne sera jamais nul.
     */
    public static LineString[] getInstances(final float[] data) {
        return getInstances(data, 0, data.length);
    }

    /**
     * Construit des objets mmorisant les coordonnes <code>data</code> de l'index
     * <code>lower</code> inclusivement jusqu' <code>upper</code> exclusivement. Ces
     * index doivent se rfrer  la position absolue dans le tableau <code>data</code>,
     * c'est--dire tre le double de l'index de la coordonne. Les valeurs <code>NaN</code>
     * au dbut et  la fin de <code>data</code> seront ignores. Celles qui apparaissent au
     * milieu auront pour effet de sparer le trait en plusieurs polylignes.
     *
     * @param data   Tableau de coordonnes (peut contenir des NaN).
     * @param lower  Index de la premire donne  considrer.
     * @param upper  Index suivant celui de la dernire donne.
     * @return       Tableau de polylignes. Peut avoir une longueur de 0, mais ne sera jamais nul.
     */
    public static LineString[] getInstances(float[] data, int lower, int upper) {
        if (REMOVE_DOUBLONS) {
            final float[] candidate = removeDoublons(data, lower, upper);
            if (candidate != null) {
                data  = candidate;
                lower = 0;
                upper = data.length;
            }
        }
        final List polylines = new ArrayList();
        for (int i=lower; i<upper; i+=2) {
            if (!Float.isNaN(data[i]) && !Float.isNaN(data[i+1])) {
                final int lowerValid = i;
                while ((i+=2) < upper) {
                    if (Float.isNaN(data[i]) || Float.isNaN(data[i+1])) {
                        break;
                    }
                }
                final PointArray points = DefaultArray.getInstance(data, lowerValid, i, true);
                if (points != null) {
                    polylines.add(new LineString(points));
                }
            }
        }
        return (LineString[]) polylines.toArray(new LineString[polylines.size()]);
    }

    /**
     * Remove consecutive identical points, since it hurt many algorithms in this
     * package. {@link Float#NaN} values are ignored (they may have doublons).
     *
     * @param  data  The data to examine.
     * @param  lower The lower index to examine in <code>data</code>, inclusive.
     * @param  upper The upper index to examine in <code>data</code>, inclusive.
     * @return <code>null</code> if no doublons was found in <code>data</code>,
     *         otherwise a new array without doublons.
     */
    private static float[] removeDoublons(final float[] data, final int lower, final int upper) {
        int dest = 0;
        float[] copy = null;
        for (int i=lower; (i+=2)<upper;) {
            if (data[i-2]==data[i] && data[i-1]==data[i+1]) {
                if (copy == null) {
                    dest = i-lower;
                    copy = new float[upper-lower-2];
                    System.arraycopy(data, lower, copy, 0, dest);
                }
                continue;
            }
            if (copy != null) {
                copy[dest++] = data[i  ];
                copy[dest++] = data[i+1];
            }
        }
        if (copy != null) {
            copy = XArray.resize(copy, dest);
        }
        return copy;
    }

    /**
     * Renvoie le premier lment de la liste  laquelle appartient la
     * polyligne. Cette mthode peut retourner <code>scan</code>, mais
     * jamais <code>null</code>  (sauf si l'argument <code>scan</code>
     * est nul).
     */
    private static LineString getFirst(LineString scan) {
        if (scan != null) {
            while (scan.previous != null) {
                scan = scan.previous;
                assert scan.previous != scan;
                assert scan.next     != scan;
            }
        }
        return scan;
    }

    /**
     * Renvoie le dernier lment de la liste  laquelle appartient la
     * polyligne. Cette mthode peut retourner <code>scan</code>, mais
     * jamais <code>null</code>  (sauf si l'argument <code>scan</code>
     * est nul).
     */
    private static LineString getLast(LineString scan) {
        if (scan != null) {
            while (scan.next != null) {
                scan = scan.next;
                assert scan.previous != scan;
                assert scan.next     != scan;
            }
        }
        return scan;
    }

    /**
     * Ajoute la polyligne <code>toAdd</code>  la fin de la polyligne <code>queue</code>.
     * Les arguments <code>queue</code> et <code>toAdd</code> peuvent tre n'importe
     * quel maillon d'une chane, mais cette mthode sera plus rapide si <code>queue</code>
     * est le dernier maillon.
     *
     * @param  queue <code>LineString</code>  la fin duquel ajouter <code>toAdd</code>. Si cet
     *               argument est nul, alors cette mthode retourne directement <code>toAdd</code>.
     * @param  toAdd <code>LineString</code>  ajouter  <code>queue</code>. Cet objet sera ajout
     *               mme s'il est vide. Si cet argument est nul, alors cette mthode retourne
     *               <code>queue</code> sans rien faire.
     * @return <code>LineString</code> rsultant de la fusion. Les anciens objets <code>queue</code>
     *         et <code>toAdd</code> peuvent avoir t modifis et ne devraient plus tre utiliss.
     * @throws IllegalArgumentException si <code>toAdd</code> avait dj t ajout 
     *         <code>queue</code>.
     */
    public static LineString append(LineString queue, LineString toAdd)
            throws IllegalArgumentException
    {
        // On doit faire l'ajout mme si 'toAdd' est vide.
        final LineString veryLast = getLast(toAdd);
        toAdd = getFirst(toAdd);
        queue = getLast (queue);
        if (toAdd == null) return queue;
        if (queue == null) return toAdd;
        if (queue == veryLast) {
            throw new IllegalArgumentException();
        }

        assert queue.next     == null;
        assert toAdd.previous == null;
        queue.next     = toAdd;
        toAdd.previous = queue;

        assert getFirst(queue) == getFirst(toAdd);
        assert getLast (queue) == getLast (toAdd);
        assert veryLast.next   == null;
        return veryLast;
    }

    /**
     * Supprime ce maillon de la chane. Ce maillon
     * conservera toutefois ses donnes.
     */
    private void remove() {
        if (previous != null) {
            previous.next = next;
        }
        if (next != null) {
            next.previous = previous;
        }
        previous = next = null;
    }

    /**
     * Indique si cette polyligne est vide. Une polyligne est vide si tous
     * ces tableaux sont nuls. Cette mthode ne vrifie pas l'tat des
     * autres maillons de la chane.
     */
    private boolean isEmpty() {
        return array==null && suffix==null;
    }

    /**
     * Retourne un des tableaux de donnes de cette polyligne. Le tableau retourn
     * peut tre {@link #prefix}, {@link #array} ou {@link #suffix} selon que
     * l'argument est -1, 0 ou +1 respectivement. Toute autre valeur lancera
     * une exception.
     *
     * @param arrayID Un code compris entre {@link #FIRST_ARRAY}
     *                et {@link #LAST_ARRAY} inclusivement.
     */
    private PointArray getArray(final int arrayID) {
        switch (arrayID) {
        //  case -1: return prefix;
            case  0: return array;
            case +1: return suffix;
            default: throw new IllegalArgumentException(String.valueOf(arrayID));
        }
    }

    /**
     * Modifie un des tableaux de donnes de cette polyligne. Le tableau modifi
     * peut tre {@link #prefix}, {@link #array} ou {@link #suffix} selon que
     * l'argument est -1, 0 ou +1 respectivement.  Toute autre valeur lancera
     * une exception.
     *
     * @param arrayID Un code compris entre {@link #FIRST_ARRAY}
     *                et {@link #LAST_ARRAY} inclusivement.
     */
    private void setArray(final int arrayID, final PointArray data) {
        switch (arrayID) {
        //  case -1: prefix=data; break;
            case  0: array =data; break;
            case +1: suffix=data; break;
            default: throw new IllegalArgumentException(String.valueOf(arrayID));
        }
    }

    /**
     * Returns an estimation of memory usage in bytes.  This method is for information
     * purpose only. The memory really used by two polylines may be lower than the sum
     * of their  <code>getMemoryUsage()</code>  return values,  since polylines try to
     * share their data when possible. Furthermore, this method do not take in account
     * the extra bytes generated by Java Virtual Machine for each objects.
     *
     * @return An <em>estimation</em> of memory usage in bytes.
     */
    public static long getMemoryUsage(LineString scan) {
        scan = getFirst(scan);
        long count = 16; // Take in account 4 internal fields of reference type (4 bytes each).
        while (scan != null) {
            for (int i=FIRST_ARRAY; i<=LAST_ARRAY; i++) {
                final PointArray data = scan.getArray(i);
                if (data!=null) {
                    count += data.getMemoryUsage();
                }
            }
            scan = scan.next;
        }
        return count;
    }

    /**
     * Retourne le nombre de points de la polyligne spcifie
     * ainsi que de tous les polylignes qui le suivent.
     *
     * @param scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *             mais cette mthode sera plus rapide si c'est le premier maillon.
     */
    public static int getPointCount(LineString scan) {
        scan = getFirst(scan);
        int count = 0;
        while (scan != null) {
            for (int i=FIRST_ARRAY; i<=LAST_ARRAY; i++) {
                final PointArray data = scan.getArray(i);
                if (data!=null) {
                    count += data.count();
                }
            }
            scan = scan.next;
        }
        return count;
    }

    /**
     * Vrifie si la polyligne spcifie est vide.
     *
     * @param scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *             mais cette mthode sera plus rapide si c'est le premier maillon.
     */
    public static boolean isEmpty(LineString scan) {
        scan = getFirst(scan);
        while (scan != null) {
            if (!scan.isEmpty()) {
                return false;
            }
            scan = scan.next;
        }
        return true;
    }

    /**
     * Returns <code>true</code> if at least one point of the specified polyline is a border.
     */
    public static boolean hasBorder(LineString scan) {
        scan = getFirst(scan);
        while (scan != null) {
            if (scan.suffix != null) {
                return true;
            }
            scan = scan.next;
        }
        return false;
    }

    /**
     * Donne  la coordonne spcifie la valeur du premier point. Si une bordure a t
     * ajoute avec la mthode {@link #prepend}, elle sera pris en compte. Si cet objet
     * <code>LineString</code> ne contient aucun point, l'objet qui suit dans la chane
     * sera automatiquement interrog.
     *
     * @param  scan  Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *               mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param  point Point dans lequel mmoriser la coordonne.
     * @return L'argument <code>point</code>, ou un nouveau point
     *         si <code>point</code> tait nul.
     * @throws NoSuchElementException Si <code>scan</code> est nul
     *         ou s'il ne reste plus de points dans la chane.
     *
     * @see #getFirstPoints
     * @see #getLastPoint
     */
    public static Point2D getFirstPoint(LineString scan, final Point2D point)
            throws NoSuchElementException
    {
        scan = getFirst(scan);
        while (scan != null) {
            for (int i=FIRST_ARRAY; i<=LAST_ARRAY; i++) {
                final PointArray data=scan.getArray(i);
                if (data != null) {
                    return data.getFirstPoint(point);
                }
            }
            scan = scan.next;
        }
        throw new NoSuchElementException();
    }

    /**
     * Donne  la coordonne spcifie la valeur du dernier point. Si une bordure a t
     * ajoute avec la mthode {@link #append}, elle sera pris en compte.  Si cet objet
     * <code>LineString</code> ne contient aucun point, l'objet qui prcde dans la chane
     * sera automatiquement interrog.
     *
     * @param  scan  Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *               mais cette mthode sera plus rapide si c'est le dernier maillon.
     * @param  point Point dans lequel mmoriser la coordonne.
     * @return L'argument <code>point</code>, ou un nouveau point
     *         si <code>point</code> tait nul.
     * @throws NoSuchElementException Si <code>scan</code> est nul
     *         ou s'il ne reste plus de points dans la chane.
     *
     * @see #getLastPoints
     * @see #getFirstPoint
     */
    public static Point2D getLastPoint(LineString scan, final Point2D point)
            throws NoSuchElementException
    {
        scan = getLast(scan);
        while (scan != null) {
            for (int i=LAST_ARRAY; i>=FIRST_ARRAY; i--) {
                PointArray data=scan.getArray(i);
                if (data != null) {
                    return data.getLastPoint(point);
                }
            }
            scan = scan.previous;
        }
        throw new NoSuchElementException();
    }

    /**
     * Donne aux coordonnes spcifies les valeurs des premiers points.
     *
     * @param scan   Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *               mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param points Tableau dans lequel mmoriser les premires coordonnes. <code>points[0]</code>
     *               contiendra la premire coordonne, <code>points[1]</code> la seconde, etc. Si
     *               un lment de ce tableau est nul, un objet {@link Point2D} sera automatiquement
     *               cr.
     *
     * @throws NoSuchElementException Si <code>scan</code> est nul ou
     *         s'il ne reste pas suffisament de points dans la chane.
     */
    public static void getFirstPoints(LineString scan, final Point2D points[])
            throws NoSuchElementException
    {
        scan = getFirst(scan);
        if (points.length == 0) {
            return;
        }
        if (scan == null) {
            throw new NoSuchElementException();
        }
        int      arrayID = FIRST_ARRAY;
        PointArray  data = null;
        PointIterator it = null;
        for (int j=0; j<points.length; j++) {
            while (it==null || !it.hasNext()) {
                if (arrayID > LAST_ARRAY) {
                    arrayID = FIRST_ARRAY;
                    scan    = scan.next;
                    if (scan == null) {
                        throw new NoSuchElementException();
                    }
                }
                data = scan.getArray(arrayID++);
                if (data != null) {
                    it = data.iterator(0);
                }
            }
            if (points[j] == null) {
                points[j] = new Point2D.Float(it.nextX(), it.nextY());
            } else {
                points[j].setLocation(it.nextX(), it.nextY());
            }
            if (REMOVE_DOUBLONS) {
                assert j==0 || !points[j].equals(points[j-1]) : scan;
            }
        }
        assert Utilities.equals(getFirstPoint(scan, null), points[0]) : scan;
    }

    /**
     * Donne aux coordonnes spcifies les valeurs des derniers points.
     *
     * @param scan   Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *               mais cette mthode sera plus rapide si c'est le dernier maillon.
     * @param points Tableau dans lequel mmoriser les dernires coordonnes.
     *               <code>points[length-1]</code> contiendra la dernire coordonne,
     *               <code>points[length-2]</code> l'avant dernire, etc. Si un lment de
     *               ce tableau est nul, un objet {@link Point2D} sera automatiquement cr.
     *
     * @throws NoSuchElementException Si <code>scan</code> est nul ou
     *         s'il ne reste pas suffisament de points dans la chane.
     */
    public static void getLastPoints(LineString scan, final Point2D points[])
            throws NoSuchElementException
    {
        scan = getLast(scan);
        if (points.length == 0) {
            // Ncessaire pour l'implmentation ci-dessous.
            return;
        }
        if (scan == null) {
            throw new NoSuchElementException();
        }
        int startIndex = -points.length;
        int    arrayID = LAST_ARRAY+1;
        PointArray data;
        /*
         * Recherche la position  partir d'o lire les donnes.  A la
         * sortie de cette boucle, la premire donne valide sera  la
         * position <code>scan.getArray(arrayID).iterator(i)</code>.
         */
        do {
            do {
                if (--arrayID < FIRST_ARRAY) {
                    arrayID = LAST_ARRAY;
                    scan = scan.previous;
                    if (scan==null) {
                        throw new NoSuchElementException();
                    }
                }
                data = scan.getArray(arrayID);
            }
            while (data==null);
            startIndex += data.count();
        }
        while (startIndex < 0);
        /*
         * Procde  la mmorisation des coordonnes.   Note: parvenu  ce stade, 'data' devrait
         * obligatoirement tre non-nul. Un {@link NullPointerException} dans le code ci-dessous
         * serait une erreur de programmation.
         */
        PointIterator it = data.iterator(startIndex);
        for (int j=0; j<points.length; j++) {
            while (!it.hasNext()) {
                do {
                    if (++arrayID > LAST_ARRAY) {
                        arrayID = FIRST_ARRAY;
                        scan = scan.next;
                    }
                    data = scan.getArray(arrayID);
                }
                while (data==null);
                it = data.iterator(0);
            }
            if (points[j] == null) {
                points[j]=new Point2D.Float(it.nextX(), it.nextY());
            } else {
                points[j].setLocation(it.nextX(), it.nextY());
            }
            if (REMOVE_DOUBLONS) {
                assert j==0 || !points[j].equals(points[j-1]) : scan;
            }
        }
        assert !it.hasNext();
        assert Utilities.equals(getLastPoint(scan, null), points[points.length-1]) : scan;
    }

    /**
     * Retourne une polyligne qui couvrira les donnes de cette polyligne
     * de l'index <code>lower</code> inclusivement jusqu' l'index
     * <code>upper</code> exclusivement.
     *
     * @param scan  Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *              mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param lower Index du premier point  retenir.
     * @param upper Index suivant celui du dernier point  retenir.
     * @return      Une chane de nouvelles polylignes, ou <code>scan</code> si aucun
     *              point n'a t ignors. Si la polyligne obtenu ne contient aucun
     *              point, alors cette mthode retourne <code>null</code>.
     */
    public static LineString subpoly(LineString scan, int lower, int upper) {
        if (lower == upper) {
            return null;
        }
        scan = getFirst(scan);
        if (lower==0 && upper==getPointCount(scan)) {
            return scan;
        }
        LineString queue=null;
        while (scan!=null) {
            LineString toAdd = null;
            for (int i=FIRST_ARRAY; i<=LAST_ARRAY; i++) {
                PointArray data = scan.getArray(i);
                if (data == null) {
                    continue;
                }
                /*
                 * Vrifie si le tableau 'data' contient au moins quelques points
                 *  prendre en compte. Si ce n'est pas le cas, il sera ignor en
                 * bloc.
                 */
                int count = data.count();
                if (count <= lower) {
                    lower -= count;
                    upper -= count;
                    continue;
                }
                /*
                 * Prend en compte les donnes de 'data' de 'lower' jusqu' 'upper',
                 * mais sans dpasser la longueur du tableau. S'il reste encore des
                 * points  aller chercher (upper!=0), on examinera les tableaux suivants.
                 */
                if (count > upper) {
                    count = upper;
                }
                assert lower >= 0 : lower;
                assert count <= data.count() : count;
                data = data.subarray(lower, count);
                if (data != null) {
                    if (toAdd == null) {
                        toAdd = new LineString(null);
                        queue = append(queue, toAdd);
                    }
                    assert toAdd.getArray(i)==null;
                    toAdd.setArray(i, data);
                }
                lower  = 0;
                upper -= count;
                if (upper==0) {
                    return queue;
                }
            }
            scan = scan.next;
        }
        throw new IndexOutOfBoundsException();
    }

    /**
     * Ajoute des points  la bordure de cette polyligne. Cette mthode est rserve
     *  un usage interne par {@link #prependBorder} et {@link #appendBorder}.
     */
    private void addBorder(float[] data, int lower, int upper, final boolean toEnd) {
        if (REMOVE_DOUBLONS) {
            final float[] candidate = removeDoublons(data, lower, upper);
            if (candidate != null) {
                data  = candidate;
                lower = 0;
                upper = data.length;
            }
        }
        if (suffix == null) {
            suffix = DefaultArray.getInstance(data, lower, upper, true);
        } else {
            suffix = suffix.insertAt(toEnd ? suffix.count() : 0, data, lower, upper, false);
        }
    }

    /**
     * Ajoute des points au dbut de cette polyligne. Ces points seront considrs comme
     * faisant partie de la bordure de la carte, et non comme des points reprsentant
     * une structure gographique.
     *
     * @param  scan  Polyligne. Cet argument peut tre n'importe quel maillon d'une chane.
     * @param  data  Coordonnes  ajouter sous forme de paires de nombres (x,y).
     * @param  lower Index du premier <var>x</var>  ajouter  la bordure.
     * @param  upper Index suivant celui du dernier <var>y</var>  ajouter  la bordure.
     * @return LineString rsultant. Ca sera en gnral <code>scan</code>.
     */
    public static LineString prependBorder(LineString scan, final float[] data,
                                           int lower, int upper)
    {
        if (REMOVE_DOUBLONS_IN_BORDER) {
            try {
                final Point2D check = getFirstPoint(scan, null);
                final float x = (float)check.getX();
                final float y = (float)check.getY();
                while (lower<upper && data[upper-2]==x && data[upper-1]==y) {
                    upper -= 2;
                }
            } catch (NoSuchElementException exception) {
                // No points in this polyline, no doublons, no problem. Continue...
            }
        }
        final int length = upper-lower;
        if (length > 0) {
            scan = getFirst(scan);
            if (scan==null || scan.array!=null) {
                scan = getFirst(append(new LineString(null), scan));
                assert scan.array==null;
            }
            scan.addBorder(data, lower, upper, false);
        }
        return scan;
    }

    /**
     * Ajoute des points  la fin de cette polyligne. Ces points seront considrs comme
     * faisant partie de la bordure de la carte, et non comme des points reprsentant
     * une structure gographique.
     *
     * @param  scan  Polyline. Cet argument peut tre n'importe quel maillon d'une chane.
     * @param  data  Coordonnes  ajouter sous forme de paires de nombres (x,y).
     * @param  lower Index du premier <var>x</var>  ajouter  la bordure.
     * @param  upper Index suivant celui du dernier <var>y</var>  ajouter  la bordure.
     * @return Polyligne rsultante. Ca sera en gnral <code>scan</code>.
     */
    public static LineString appendBorder(LineString scan, final float[] data, int lower, int upper) {
        if (REMOVE_DOUBLONS_IN_BORDER) {
            try {
                final Point2D check = getLastPoint(scan, null);
                final float x = (float)check.getX();
                final float y = (float)check.getY();
                while (lower<upper && data[lower]==x && data[lower+1]==y) {
                    lower += 2;
                }
            } catch (NoSuchElementException exception) {
                // No points in this polyline, no doublons, no problem. Continue...
            }
        }
        final int length = upper-lower;
        if (length > 0) {
            scan = getLast(scan);
            if (scan == null) {
                scan = new LineString(null);
            }
            scan.addBorder(data, lower, upper, true);
        }
        return scan;
    }

    /**
     * Inverse l'ordre de tous les points.  Cette mthode retournera le
     * premier maillon d'une nouvelle chane de polylignes qui contiendra
     * les donnes en ordre inverse.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *              mais cette mthode sera plus rapide si c'est le dernier maillon.
     */
    public static LineString reverse(LineString scan) {
        LineString queue = null;
        for (scan=getLast(scan); scan!=null; scan=scan.previous) {
            for (int arrayID=LAST_ARRAY; arrayID>=FIRST_ARRAY; arrayID--) {
                PointArray array = scan.getArray(arrayID);
                if (array != null) {
                    array = array.reverse();
                    /*
                     * Tous les tableaux sont balays dans cette boucle,
                     * un  un et dans l'ordre inverse. Les prfix doivent
                     * devenir des suffix, et les suffix doivent devenir
                     * des prfix.
                     */
                    if (arrayID == 0) {
                        queue = append(queue, new LineString(array));
                    } else {
                        queue = getLast(queue); // Par prcaution.
                        if (queue == null) {
                            queue = new LineString(null);
                        }
                        assert queue.suffix==null;
                        queue.suffix=array;
                    }
                }
            }
        }
        return queue;
    }

    /**
     * Retourne les coordonnes d'une bote qui englobe compltement tous
     * les points de la polyligne. Si cette polyligne ne contient aucun point,
     * alors cette mthode retourne <code>null</code>.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *              mais cette mthode sera plus rapide si c'est le premier maillon.
     * @return Un rectangle englobeant toutes les coordonnes de cette polyligne et de
     *         ceux qui la suivent, ou <code>null</code>.
     */
    public static Rectangle2D getBounds2D(LineString scan) {
        Rectangle2D bounds = null;
        for (scan=getFirst(scan); scan!=null; scan=scan.next) {
            for (int arrayID=FIRST_ARRAY; arrayID<=LAST_ARRAY; arrayID++) {
                final PointArray array = scan.getArray(arrayID);
                if (array != null) {
                    final Rectangle2D toAdd = array.getBounds2D();
                    if (toAdd != null) {
                        if (bounds != null) {
                            bounds.add(toAdd);
                        } else {
                            bounds = toAdd;
                        }
                    }
                }
            }
        }
        return bounds;
    }

    /**
     * Retourne les coordonnes d'une bote qui englobe compltement tous
     * les points de la polyligne. Si cette polyligne ne contient aucun point,
     * alors cette mthode retourne <code>null</code>.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *              mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param  transform Transformation  appliquer sur les donnes (nulle pour aucune).
     * @return Un rectangle englobeant toutes les coordonnes de cette polyligne et de
     *         ceux qui la suivent, ou <code>null</code>.
     * @throws TransformException Si une projection cartographique a chou.
     */
    public static Rectangle2D getBounds2D(LineString scan, final MathTransform2D transform)
            throws TransformException
    {
        if (transform==null || transform.isIdentity()) {
            return getBounds2D(scan);
        }
        float xmin = Float.POSITIVE_INFINITY;
        float xmax = Float.NEGATIVE_INFINITY;
        float ymin = Float.POSITIVE_INFINITY;
        float ymax = Float.NEGATIVE_INFINITY;
        final Point2D.Float point = new Point2D.Float();
        for (scan=getFirst(scan); scan!=null; scan=scan.next) {
            for (int arrayID=FIRST_ARRAY; arrayID<=LAST_ARRAY; arrayID++) {
                final PointArray array = scan.getArray(arrayID);
                if (array != null) {
                    final PointIterator it = array.iterator(0);
                    while (it.hasNext()) {
                        point.x = it.nextX();
                        point.y = it.nextY();
                        transform.transform(point, point);
                        if (point.x<xmin) xmin=point.x;
                        if (point.x>xmax) xmax=point.x;
                        if (point.y<ymin) ymin=point.y;
                        if (point.y>ymax) ymax=point.y;
                    }
                }
            }
        }
        if (xmin<=xmax && ymin<=ymax) {
            return new Rectangle2D.Float(xmin, ymin, xmax-xmin, ymax-ymin);
        } else {
            return null;
        }
    }

    /**
     * Renvoie des statistiques sur la rsolution d'un polyligne. Cette rsolution sera
     * la distance moyenne entre deux points du polyligne,  mais sans prendre en compte
     * les "points de bordure"  (par exemple les points qui suivent le bord d'une carte
     * plutt que de reprsenter une structure gographique relle).
     * <br><br>
     * La rsolution est calcule en utilisant le systme de coordonnes spcifi. Les
     * units du rsultat seront donc  les units des deux premiers axes de ce systme
     * de coordonnes,  <strong>sauf</strong>  si les deux premiers axes utilisent des
     * coordonnes gographiques angulaires  (c'est le cas notamment des objets {@link
     * GeographicCoordinateSystem}).  Dans ce dernier cas,  le calcul utilisera plutt
     * les distances orthodromiques sur l'ellipsode ({@link Ellipsoid}) du systme de
     * coordonnes.   En d'autres mots, pour les systmes cartographiques, le rsultat
     * de cette mthode sera toujours exprim en units linaires (souvent des mtres)
     * peu importe que le systme de coordonnes soit {@link ProjectedCoordinateSystem}
     * ou {@link GeographicCoordinateSystem}.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *         mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param  transformation Systmes de coordonnes source et destination.
     *         <code>getSourceCS()</code> doit tre le systme interne des points
     *         des polylignes, tandis que  <code>getTargetCS()</code> doit tre le
     *         systme dans lequel faire le calcul. C'est <code>getTargetCS()</code>
     *         qui dterminera les units du rsultat. Cet argument peut tre nul
     *         si aucune transformation n'est ncessaire. Dans ce cas, le systme
     *         de coordonnes <code>getTargetCS()</code> sera suppos cartsien.
     * @return Statistiques sur la rsolution. L'objet retourn ne sera jamais nul, mais les
     *         statistiques seront tous  NaN si cette courbe de niveau ne contenait aucun
     *         point. Voir la description de cette mthode pour les units.
     * @throws TransformException Si une transformation de coordonnes a choue.
     */
    static Statistics getResolution(LineString scan, final CoordinateTransformation transformation)
            throws TransformException
    {
        /*
         * Checks the coordinate system validity. If valid and if geographic,
         * gets the ellipsoid to use for orthodromic distance computations.
         */
        final MathTransform2D transform;
        final Ellipsoid       ellipsoid;
        final Unit         xUnit, yUnit;
        if (transformation != null) {
            final MathTransform tr = transformation.getMathTransform();
            transform = !tr.isIdentity() ? (MathTransform2D) tr : null;
            final CoordinateSystem targetCS = transformation.getTargetCS();
            xUnit = targetCS.getUnits(0);
            yUnit = targetCS.getUnits(1);
            if (!Utilities.equals(xUnit, yUnit)) {
                throw new IllegalArgumentException(Resources.format(
                                            ResourceKeys.ERROR_NON_CARTESIAN_COORDINATE_SYSTEM_$1,
                                            targetCS.getName(null)));
            }
            ellipsoid = CTSUtilities.getHeadGeoEllipsoid(targetCS);
        } else {
            transform = null;
            ellipsoid = null;
            xUnit = yUnit = null;
        }
        /*
         * Compute statistics...
         */
        final Statistics stats = new Statistics();
        Point2D          point = new Point2D.Double();
        Point2D           last = null;
        for (scan=getFirst(scan); scan!=null; scan=scan.next) {
            final PointArray array = scan.array;
            if (array == null) {
                continue;
            }
            final PointIterator it = array.iterator(0);
            while (it.hasNext()) {
                point.setLocation(it.nextX(), it.nextY());
                if (transform != null) {
                    point = transform.transform(point, point);
                }
                final double distance;
                if (ellipsoid != null) {
                    point.setLocation(Unit.DEGREE.convert(point.getX(), xUnit),
                                      Unit.DEGREE.convert(point.getY(), yUnit));
                    if (last == null) {
                        last = (Point2D) point.clone();
                        continue;
                    }
                    distance = ellipsoid.orthodromicDistance(last, point);
                } else {
                    if (last == null) {
                        last = (Point2D) point.clone();
                        continue;
                    }
                    distance = last.distance(point);
                }
                stats.add(distance);
                final Point2D swap = last;
                last = point;
                point = swap;
            }
        }
        return stats;
    }

    /**
     * Modifie la rsolution de cette carte. Cette mthode procdera en interpolant les donnes
     * de faon  ce que chaque point soit spar du prcdent par la distance spcifie.  Cela
     * peut se traduire par des conomies importante de mmoire  si  une trop grande rsolution
     * n'est pas ncessaire. Notez que cette opration est irreversible.  Appeler cette mthode
     * une seconde fois avec une rsolution plus fine gonflera la taille des tableaux internes,
     * mais sans amlioration relle de la prcision.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *         mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param  transformation Transformation permettant de convertir les coordonnes des polylignes
     *         vers des coordonnes cartsiennes. Cet argument peut tre nul si les coordonnes de
     *         <code>this</code> sont dj exprimes selon un systme de coordonnes cartsiennes.
     * @param  resolution Rsolution dsire, selon les mmes units que {@link #getResolution}.
     * @throws TransformException Si une erreur est survenue lors d'une projection cartographique.
     *
     * @see #getResolution
     */
    public static void setResolution(LineString scan, final CoordinateTransformation transformation,
                                     double resolution)
            throws TransformException
    {
        /*
         * Checks arguments validity. This method do not support latitude/longitude
         * coordinates. Coordinates must be projected in some linear units.
         */
        if (!(resolution > 0)) {
            throw new IllegalArgumentException(String.valueOf(resolution));
        }
        final MathTransform2D transform;
        final MathTransform2D inverseTransform;
        if (transformation != null) {
            final CoordinateSystem targetCS = transformation.getTargetCS();
            if (CTSUtilities.getHeadGeoEllipsoid(targetCS)!=null ||
                !Utilities.equals(targetCS.getUnits(0), targetCS.getUnits(1)))
            {
                throw new IllegalArgumentException(Resources.format(
                                            ResourceKeys.ERROR_NON_CARTESIAN_COORDINATE_SYSTEM_$1,
                                            targetCS.getName(null)));
            }
            final MathTransform tr = transformation.getMathTransform();
            if (!tr.isIdentity()) {
                transform        = (MathTransform2D) tr;
                inverseTransform = (MathTransform2D) transform.inverse();
            } else {
                transform        = null;
                inverseTransform = null;
            }
        } else {
            transform        = null;
            inverseTransform = null;
        }
        /*
         * Performs the linear interpolations, assuming
         * that we are using a cartesian coordinate system.
         */
        for (scan=getFirst(scan); scan!=null; scan=scan.next) {
            final PointArray points = scan.array;
            if (points == null) {
                continue;
            }
            /*
             * Obtiens les coordonnes projetes. Si ces coordonnes reprsentent des
             * degrs de longitudes et latitudes, alors une projection cartographique
             * sera obligatoire afin de faire correctement les calculs de distances.
             */
            float[] array = points.toArray();
            assert (array.length & 1)==0;
            if (transform!=null && !transform.isIdentity()) {
                transform.transform(array, 0, array, 0, array.length/2);
            }
            if (array.length >= 2) {
                /*
                 * Effectue la dcimation des coordonnes. La toute premire
                 * coordonne sera conserve inchange. Il en ira de mme de
                 * la dernire,  la fin de ce bloc.
                 */
                final Point2D.Float point = new Point2D.Float(array[0], array[1]);
                final Line2D.Float   line = new  Line2D.Float(0,0, point.x, point.y);
                int destIndex   = 2; // Ne touche pas au premier point.
                int sourceIndex = 2; // Le premier point est dj lu.
                while (sourceIndex < array.length) {
                    line.x1 = line.x2;
                    line.y1 = line.y2;
                    line.x2 = array[sourceIndex++];
                    line.y2 = array[sourceIndex++];
                    Point2D next;
                    while ((next=Geometry.colinearPoint(line, point, resolution)) != null) {
                        if (destIndex == sourceIndex) {
                            final int extra = 256;
                            final float[] oldArray=array;
                            array=new float[array.length + extra];
                            System.arraycopy(oldArray, 0,         array, 0,                                  destIndex);
                            System.arraycopy(oldArray, destIndex, array, sourceIndex+=extra, oldArray.length-destIndex);
                        }
                        assert destIndex < sourceIndex;
                        array[destIndex++] = line.x1 = point.x = (float)next.getX();
                        array[destIndex++] = line.y1 = point.y = (float)next.getY();
                    }
                }
                /*
                 * La dcimation est maintenant termine. Vrifie si le dernier point
                 * apparat dans le tableau dcim. S'il n'apparat pas, on l'ajoutera.
                 * Ensuite, on librera la mmoire rserve en trop.
                 */
                if (array[destIndex-2] != line.x2  ||  array[destIndex-1] != line.y2) {
                    if (destIndex == array.length) {
                        array = XArray.resize(array, destIndex+2);
                    }
                    array[destIndex++] = line.x2;
                    array[destIndex++] = line.y2;
                }
                if (destIndex != array.length) {
                    array = XArray.resize(array, destIndex);
                }
            }
            /*
             * Les interpolations tant termines, reconvertit les coordonnes
             * selon leur systme de coordonns initial et mmorise le nouveau
             * tableau dcim  la place de l'ancien.
             */
            if (inverseTransform != null) {
                inverseTransform.transform(array, 0, array, 0, array.length/2);
            }
            scan.array = (array!=null && array.length!=0) ? new DefaultArray(array) : null;
        }
    }

    /**
     * Dclare que les donnes de cette polyligne ne vont plus changer. Cette
     * mthode peut raranger les tableaux de points d'une faon plus compacte.
     *
     * @param  scan     Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *                  mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param  close    <code>true</code> pour indiquer que ces polylignes reprsentent une
     *                  forme gomtrique ferme (donc un polygone).
     * @param  compress An optional compression level, or <code>null</code> to lets the compression
     *                  unchanged.
     *
     * @return La polyligne compresse (habituellement <code>scan</code> lui-mme),
     *         ou <code>null</code> si la polyligne ne contenait aucune donne.
     */
    public static LineString freeze(LineString scan, final boolean close,
                                    final CompressionLevel compress) {
        scan = getFirst(scan);
        /*
         * Etape 1: Si on a demand  fermer le polygone, vrifie si le premier maillon de
         *          la chane ne contenait qu'une bordure.  Si c'est le cas, on dmnagera
         *          cette bordure  la fin du dernier maillon.
         */
        if (close && scan!=null && scan.suffix!=null && scan.array==null) {
            LineString last = getLast(scan);
            if (last != scan) {
                if (last.suffix != null) {
                    last.suffix = last.suffix.insertAt(last.suffix.count(), scan.suffix, false);
                } else {
                    last.suffix = scan.suffix;
                }
                scan.suffix = null;
            }
        }
        /*
         * Etape 2: Fusionne ensemble des polylignes qui peuvent l'tre.
         *          Deux polylignes peuvent tre fusionnes ensemble si elles
         *          ne sont spares par aucune bordure, ou si elle sont toutes
         *          deux des bordures.
         */
        if (scan != null) {
            LineString previous = scan;
            LineString current  = scan;
            while ((current=current.next) != null) {
                if (previous.suffix == null) {
                    if (previous.array != null) {
                        // Dmnage le tableau de points de 'previous' au dbut
                        // de celui de 'current' si aucune bordure ne les spare.
                        if (current.array != null) {
                            if (MERGE_POLYLINE_DATA) {
                                current.array = current.array.insertAt(0, previous.array, false);
                                previous.array = null;
                            }
                        } else {
                            current.array = previous.array;
                            previous.array = null;
                        }
                    }
                } else {
                    if (current.array == null) {
                        // Dmnage le suffix de 'previous' au dbut de
                        // celui de 'current' si rien ne les spare.
                        if (current.suffix != null) {
                            current.suffix = current.suffix.insertAt(0, previous.suffix, false);
                        } else {
                            current.suffix = previous.suffix;
                        }
                        previous.suffix = null;
                    }
                }
                previous=current;
            }
        }
        /*
         * Etape 3: Gle et compresse les tableaux de points, et
         *          limine les ventuels tableaux devenus inutile.
         */
        LineString root = scan;
        while (scan!=null) {
            /*
             * Comprime tous les tableaux d'un maillon de la chane.
             * La compression maximale ("full") ne sera toutefois pas
             * applique sur les "points de bordure".
             */
            for (int arrayID=FIRST_ARRAY; arrayID<=LAST_ARRAY; arrayID++) {
                final PointArray array = scan.getArray(arrayID);
                if (array != null) {
                    scan.setArray(arrayID, array.getFinal(arrayID==0 ? compress : null));
                }
            }
            /*
             * Supprime les maillons devenus vides. Ca peut avoir pour effet
             * de changer de maillon ("root") pour le dbut de la chane.
             */
            LineString current=scan;
            scan = scan.next;
            if (current.isEmpty()) {
                current.remove();
                if (current == root) {
                    root = scan;
                }
            }
        }
        return root;
    }

    /**
     * Copy (<var>x</var>,<var>y</var>) coordinates in the specified destination array.
     * If <code>resolution</code> is greater than 0, then points that are closer than
     * <code>resolution</code> from previous one will be skiped.
     *
     * @param  The destination array. The coordinates will be filled in {@link ArrayData#array}
     *         from index {@link ArrayData#length}. The array will be expanded if needed, and
     *         {@link ArrayData#length} will be updated with index after the <code>array</code>'s
     *         element filled with the last <var>y</var> ordinates.
     * @param  resolution The minimum distance desired between points, in this polyline's
     *         coordinate system.
     * @param  transform The transform to apply, or <code>null</code> if none.
     * @throws TransformException if a transformation failed.
     */
    public static void toArray(LineString poly, final ArrayData dest, float resolution,
                               final MathTransform2D transform) throws TransformException
    {
        resolution *= resolution;
        poly = getFirst(poly);
        GeneralPath path = null;
        for (LineString scan=poly; scan!=null; scan=scan.next) {
            for (int i=FIRST_ARRAY; i<=LAST_ARRAY; i++) {
                final PointArray array = scan.getArray(i);
                if (array != null) {
                    final int lower = dest.length();
                    // On ne dcime pas les points de bordure (i!=0).
                    array.toArray(dest, (i==0) ? resolution : 0);
                    if (transform != null) {
                        if (i==0) {
                            // Transform the main data: fast way, no curves.
                            final float[] data = dest.array();
                            transform.transform(data, lower, data, lower, (dest.length()-lower)/2);
                        } else {
                            // Transform the borders: slower, can create curves.
                            if (path == null) {
                                path = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
                            }
                            path.reset();
                            dest.extract(lower, path);
                            dest.append(transform.createTransformedShape(path));
                        }
                    }
                }
            }
        }
    }

    /**
     * Retourne une reprsentation de cet objet sous forme
     * de chane de caractres.  Cette reprsentation sera
     * de la forme <code>"LineString[3 of 4; 47 pts]"</code>.
     */
    public String toString() {
        final StringBuffer buffer = new StringBuffer(Utilities.getShortClassName(this));
        buffer.append('[');
        int index=1;
        for (LineString scan=previous; scan!=null; scan=scan.previous) {
            index++;
        }
        buffer.append(index);
        for (LineString scan=next; scan!=null; scan=scan.next) {
            index++;
        }
        buffer.append(" of ");
        buffer.append(index);
        buffer.append("; ");
        buffer.append(array!=null ? array.count() : 0);
        buffer.append(" points");
        if (suffix != null) {
            buffer.append(" + ");
            buffer.append(suffix.count());
            buffer.append(" in border");
        }
        buffer.append(']');
        return buffer.toString();
    }

    /**
     * Retourne un code reprsentant la polyligne spcifie.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *              mais cette mthode sera plus rapide si c'est le premier maillon.
     * @return Un code calcul  partir de quelques points de la polyligne spcifie.
     */
    public static int hashCode(LineString scan) {
        int code = (int)serialVersionUID;
        for (scan=getFirst(scan); scan!=null; scan=scan.next) {
            if (scan.array != null) {
                code = 37*code + scan.array.hashCode();
            }
        }
        return code;
    }

    /**
     * Indique si deux polylignes contiennent les mmes points. Cette mthode
     * retourne aussi <code>true</code> si les deux arguments sont nuls.
     *
     * @param poly1 Premire polyligne. Cet argument peut tre n'importe quel maillon d'une
     *              chane, mais cette mthode sera plus rapide si c'est le premier maillon.
     * @param poly2 Seconde polyligne. Cet argument peut tre n'importe quel maillon d'une
     *              chane, mais cette mthode sera plus rapide si c'est le premier maillon.
     */
    public static boolean equals(LineString poly1, LineString poly2) {
        poly1 = getFirst(poly1);
        poly2 = getFirst(poly2);
        while (poly1 != poly2) {
            if (poly1==null || poly2==null) {
                return false;
            }
            for (int arrayID=FIRST_ARRAY; arrayID<=LAST_ARRAY; arrayID++) {
                final PointArray array1 = poly1.getArray(arrayID);
                final PointArray array2 = poly2.getArray(arrayID);
                if (!Utilities.equals(array1, array2)) {
                    return false;
                }
            }
            poly1 = poly1.next;
            poly2 = poly2.next;
        }
        return true;
    }

    /**
     * Retourne une copie de la polyligne spcifie. Cette mthode ne copie que les rfrences
     * vers une version immutable des tableaux de points. Les points eux-mmes ne sont pas
     * copis, ce qui permet d'viter de consommer une quantit excessive de mmoire.
     *
     * @param  scan Polyligne. Cet argument peut tre n'importe quel maillon d'une chane,
     *              mais cette mthode sera plus rapide si c'est le premier maillon.
     * @return Copie de la chane <code>scan</code>.
     */
    public static LineString clone(LineString scan) {
        LineString queue=null;
        for (scan=getFirst(scan); scan!=null; scan=scan.next) {
            final LineString toMerge = new LineString(null);
            for (int arrayID=FIRST_ARRAY; arrayID<=LAST_ARRAY; arrayID++) {
                PointArray array = scan.getArray(arrayID);
                if (array != null) {
                    array = array.getFinal(null);
                }
                toMerge.setArray(arrayID, array);
            }
            if (!toMerge.isEmpty()) {
                queue = append(queue, toMerge);
            }
        }
        return queue;
    }




    /**
     * A set of points ({@link Point2D}) from a polyline or a polygon.
     * This set of points is returned by {@link Polyline#getPoints}.
     *
     * @version $Id: LineString.java,v 1.2 2003/07/23 14:17:32 desruisseaux Exp $
     * @author Martin Desruisseaux
     */
    static final class Collection extends AbstractCollection {
        /**
         * Premire polyligne de la chane de points  balayer.
         */
        private final LineString data;

        /**
         * Transformation  appliquer sur chacun des points.
         */
        private final MathTransform2D transform;

        /**
         * Construit un ensemble de points.
         */
        public Collection(final LineString data, final MathTransform2D transform) {
            this.data = data;
            this.transform = transform;
        }

        /**
         * Retourne le nombre de points dans cet ensemble.
         */
        public int size() {
            return getPointCount(data);
        }

        /**
         * Retourne un itrateur balayant les points de cet ensemble.
         */
        public java.util.Iterator iterator() {
            return new Iterator(data, transform);
        }
    }




    /**
     * Iterateur balayant les coordonnes d'un polyligne ou d'un polygone.
     *
     * @version $Id: LineString.java,v 1.2 2003/07/23 14:17:32 desruisseaux Exp $
     * @author Martin Desruisseaux
     */
    static final class Iterator implements java.util.Iterator {
        /**
         * Polyligne qui sert de point de dpart  cet itrateur.
         * Cette informations est utilise par {@link #rewind}.
         */
        private final LineString start;

        /**
         * Polyligne qui sera balaye par les prochains appels de {@link #next}.
         * Ce champs sera mis  jour au fur et  mesure que l'on passera d'une
         * polyligne  l'autre.
         */
        private LineString current;

        /**
         * Code indiquant quel champs de {@link #current} est prsentement en cours d'examen:
         *
         *    -1 pour {@link LineString#prefix},
         *     0 pour {@link LineString#array} et
         *    +1 pour {@link LineString#suffix}.
         */
        private int arrayID = FIRST_ARRAY-1;;

        /**
         * Itrateur balayant les donnes. Cet itrateur
         * aura t obtenu d'un tableau {@link PointArray}.
         */
        private PointIterator iterator;

        /**
         * Transformation  appliquer sur les coordonnes,
         * ou <code>null</code> s'il n'y en a pas.
         */
        private final MathTransform2D transform;

        /**
         * Point utilis temporairement pour les projections.
         */
        private final Point2D.Float point = new Point2D.Float();

        /**
         * Initialise l'itrateur de faon  dmarrer
         * les balayages  partir de la polyligne spcifie.
         *
         * @param start Polyligne (peut tre nul).
         * @param transform Transformation  appliquer sur les
         *        coordonnes, ou <code>null</code> s'il n'y en a pas.
         */
        public Iterator(final LineString start, final MathTransform2D transform) {
            this.start = current = getFirst(start);
            this.transform = (transform!=null && !transform.isIdentity()) ? transform : null;
            nextArray();
        }

        /**
         * Avance l'itrateur au prochain tableau.
         */
        private void nextArray() {
            while (current != null) {
                while (++arrayID <= LAST_ARRAY) {
                    final PointArray array = current.getArray(arrayID);
                    if (array != null) {
                        iterator = array.iterator(0);
                        if (iterator.hasNext()) {
                            return;
                        }
                    }
                }
                arrayID = LineString.FIRST_ARRAY-1;
                current = current.next;
            }
            iterator = null;
        }

        /**
         * Indique s'il reste des donnes que peut retourner {@link #next}.
         */
        public boolean hasNext() {
            while (iterator != null) {
                if (iterator.hasNext()) {
                    return true;
                }
                nextArray();
            }
            return false;
        }

        /**
         * Retourne les coordonnes du point suivant.
         *
         * @return Le point suivant comme un objet {@link Point2D}.
         */
        public Object next() throws NoSuchElementException {
            if (hasNext()) {
                Point2D point = (Point2D) iterator.next();
                if (transform != null) try {
                    point = transform.transform(point, point);
                } catch (TransformException exception) {
                    // Should not happen, since {@link Polyline#setCoordinateSystem}
                    // has already successfully projected every points.
                    unexpectedException("LineString", "next", exception);
                    return null;
                }
                return point;
            } else {
                throw new NoSuchElementException();
            }
        }

        /**
         * Retourne les coordonnes du point suivant. Contrairement  la mthode {@link #next()},
         * celle-ci retourne <code>null</code> sans lancer d'exception s'il ne reste plus de point
         *  balayer.
         *
         * @param  dest Point dans lequel mmoriser le rsultat. Si cet argument
         *         est nul, un nouvel objet sera cr et retourn pour mmoriser
         *         les coordonnes.
         * @return S'il restait des coordonnes  lire, le point <code>point</code> qui avait t
         *         spcifi en argument. Si <code>point</code> tait nul, un objet {@link Point2D}
         *         nouvellement cr. S'il ne restait plus de donnes  lire, cette mthode retourne
         *         toujours <code>null</code>.
         */
        final Point2D.Float next(Point2D.Float dest) {
            while (hasNext()) {
                if (dest != null) {
                    dest.x = iterator.nextX();
                    dest.y = iterator.nextY();
                } else {
                    dest = new Point2D.Float(iterator.nextX(), iterator.nextY());
                }
                if (transform != null) try {
                    transform.transform(dest, dest);
                } catch (TransformException exception) {
                    // Should not happen, since {@link Polyline#setCoordinateSystem}
                    // has already successfully projected every points.
                    unexpectedException("LineString", "next", exception);
                    continue;
                }
                return dest;
            }
            return null;
        }

        /**
         * Retourne les coordonnes du prochain point dans le champs
         * (<var>x2</var>,<var>y2</var>) de la ligne spcifie. Les
         * anciennes coordonnes (<var>x2</var>,<var>y2</var>) seront
         * pralablement copies dans (<var>x1</var>,<var>y1</var>).
         * Si cette mthode a russie, elle retourne <code>true</code>.
         *
         * Si elle a choue parce qu'il ne restait plus de points disponibles, elle
         * aura tout de mme copi les coordonnes (<var>x2</var>,<var>y2</var>) dans
         * (<var>x1</var>,<var>y1</var>) (ce qui aura pour effet de donner  la ligne
         * une longueur de 0) et retournera <code>false</code>.
         */
        final boolean next(final Line2D.Float line) {
            line.x1 = line.x2;
            line.y1 = line.y2;
            while (hasNext()) {
                if (transform == null) {
                    line.x2 = iterator.nextX();
                    line.y2 = iterator.nextY();
                } else try {
                    point.x = iterator.nextX();
                    point.y = iterator.nextY();
                    transform.transform(point, point);
                    line.x2 = point.x;
                    line.y2 = point.y;
                } catch (TransformException exception) {
                    // Should not happen, since {@link Polyline#setCoordinateSystem}
                    // has already successfully projected every points.
                    unexpectedException("LineString", "next", exception);
                    continue;
                }
                return true;
            }
            return false;
        }

        /**
         * Repositionne cet itrateur  son point de dpart.
         */
        final void rewind() {
            current  = start;
            arrayID  = FIRST_ARRAY-1;
            nextArray();
        }

        /**
         * Cette opration n'est pas supporte.
         *
         * @throws UnsupportedOperationException Systmatiquement lance.
         */
        public void remove() throws UnsupportedOperationException {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Mthode appele lorsqu'une erreur inatendue est survenue.
     *
     * @param source Nom de la classe dans laquelle est survenu l'exception.
     * @param method Nom de la mthode dans laquelle est survenu l'exception.
     * @param exception L'exception survenue.
     */
    static void unexpectedException(final String classe, final String method,
                                    final TransformException exception)
    {
        Utilities.unexpectedException("org.geotools.renderer.geom", classe, method, exception);
    }
}
