/*
 * Geotools 2 - OpenSource mapping toolkit
 * (C) 2003, Geotools Project Managment Committee (PMC)
 * (C) 2001, Institut de Recherche pour le Dveloppement
 * (C) 1998, Pches et Ocans Canada
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Contacts:
 *     UNITED KINGDOM: James Macgill
 *             mailto:j.macgill@geog.leeds.ac.uk
 *
 *     FRANCE: Surveillance de l'Environnement Assiste par Satellite
 *             Institut de Recherche pour le Dveloppement / US-Espace
 *             mailto:seasnet@teledetection.fr
 *
 *     CANADA: Observatoire du Saint-Laurent
 *             Institut Maurice-Lamontagne
 *             mailto:osl@osl.gc.ca
 */
package org.geotools.renderer.j2d;

// J2SE dependencies
import java.util.Locale;

// Geotools dependencies
import org.geotools.pt.CoordinateFormat;
import org.geotools.pt.CoordinatePoint;
import org.geotools.cs.CoordinateSystem;
import org.geotools.ct.TransformException;
import org.geotools.cs.GeographicCoordinateSystem;


/**
 * Formateurs des coordonnes pointes par le curseur de la souris. Les instances de cette classe
 * pourront crire les coordonnes pointes ainsi qu'une ventuelle valeurs sous cette coordonnes
 * (par exemple la temprature sur une image satellitaire de temprature).
 *
 * The {@linkplain #getCoordinateSystem output coordinate system} may have an arbitrary
 * number of dimensions (as long as a transform exists from the two-dimensional
 * {@linkplain Renderer#getCoordinateSystem renderer's coordinate system}), but
 * is usually two-dimensional.
 *
 * @version $Id: MouseCoordinateFormat.java,v 1.7 2003/05/13 11:00:47 desruisseaux Exp $
 * @author Martin Desruisseaux
 */
public class MouseCoordinateFormat extends CoordinateFormat {
    /**
     * The coordinate point to format.
     */
    private CoordinatePoint point;

    /**
     * Buffer pour l'criture des coordonnes.
     */
    private final StringBuffer buffer = new StringBuffer();

    /**
     * Indique si la mthode {@link #format} doit crire la valeur aprs la coordonne. Les
     * valeurs sont obtenues en appelant la mthode <code>RenderedLayer.formatValue(...)</code>.
     * Par dfaut, les valeurs (si elles sont disponibles) sont crites.
     */
    private boolean valueVisible = true;

    /**
     * Construct a coordinate format for the default locale.
     */
    public MouseCoordinateFormat() {
        super();
    }
    

    /**
     * Construit un objet qui crira les coordonnes pointes par le
     * curseur de la souris. Les coordonnes seront crites selon le
     * systme de coordonnes par dfaut "WGS 1984".
     *
     * @param locale The locale for formatting coordinates and numbers.
     */
    public MouseCoordinateFormat(final Locale locale) {
        super(locale);
    }

    /**
     * Indique si la mthode {@link #format} doit crire la valeur aprs la coordonne.
     * Par dfaut, les valeurs (si elles sont disponibles) sont crites.
     */
    public boolean isValueVisible() {
        return valueVisible;
    }

    /**
     * Spcifie si la mthode {@link #format} doit aussi crire la valeur aprs la
     * coordonne.
     */
    public void setValueVisible(final boolean valueVisible) {
        this.valueVisible = valueVisible;
    }

    /**
     * Retourne une chane de caractres reprsentant les coordonnes pointes par le curseur
     * de la souris.  Les coordonnes seront crites selon le systme de coordonnes spcifi
     * lors du dernier appel de {@link #setCoordinateSystem}. Si une des couches peut ajouter
     * une valeur  la coordonne (par exemple une couche qui reprsente une image satellitaire
     * de temprature) et que l'criture des valeurs est autorise (voir {@link #isValueVisible}),
     * alors la valeur sera crite aprs les coordonnes.
     *
     * @param  event Evnements contenant les coordonnes de la souris.
     * @return Chane de caractres reprsentant les coordonnes pointes
     *         par le curseur de la souris.
     */
    public String format(final GeoMouseEvent event) {
        final Renderer renderer = event.context.renderer;
        try {
            point = event.getCoordinate(getCoordinateSystem(), point);
        } catch (TransformException exception) {
            return "ERROR";
        }
        buffer.setLength(0);
        format(point, buffer, null);
        if (valueVisible) {
            final int length = buffer.length();
            buffer.append("  (");
            if (renderer.formatValue(event, buffer)) {
                buffer.append(')');
            } else {
                buffer.setLength(length);
            }
        }
        return buffer.toString();
    }
}
