/*
 * Units - Temporary implementation for Geotools 2
 * Copyright (C) 1998 University Corporation for Atmospheric Research (Unidata)
 *               1998 Bill Hibbard & al. (VisAD)
 *               1999 Pches et Ocans Canada
 *               2000 Institut de Recherche pour le Dveloppement
 *               2002 Centre for Computational Geography
 *
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *    Library General Public License for more details (http://www.gnu.org/).
 *
 *
 *    This package is inspired from the units package of VisAD.
 *    Unidata and Visad's work is fully acknowledged here.
 *
 *                   THIS IS A TEMPORARY CLASS
 *
 *    This is a placeholder for future <code>Unit</code> class.
 *    This skeleton will be removed when the real classes from
 *    JSR-108: Units specification will be publicly available.
 */
package org.geotools.units;

// Entrs/sorties
import java.io.ObjectStreamException;


/**
 * Classe reprsentant les units fondamentales d'un systme d'units.
 * Le systme international (SI) n'a que quelques units fondamentales,  partir desquelles
 * sont construites toutes les autres units. Les units fondamentales sont du systme SI sont:
 *
 * <ul>
 *     <li>Les Ampres pour le courant lectrique</li>
 *     <li>Les Candelas pour l'intensit lumineuse</li>
 *     <li>Les degrs Kelvin pour la temprature</li>
 *     <li>Les kilogrammes pour la masse</li>
 *     <li>Les mtres pour la longueur</li>
 *     <li>Les secondes pour le temps</li>
 * </ul>
 *
 * Il existe aussi d'autres mesures qui sont sans units, mais pour lesquelles
 * il est pratique de leur affecter des pseudo-units fondamentales. Les radians, les moles et
 * les mesures de salinit en sont des exemples.
 *
 * @version 1.0
 * @author Steven R. Emmerson
 * @author Bill Hibbard
 * @author Martin Desruisseaux
 */
/*public*/ 
final class BaseUnit extends SimpleUnit {
    /**
     * Serial number for interoperability with different versions.
     */
    private static final long serialVersionUID = -2736303035387288589L;
    
    /**
     * Construit une unit fondamentale.
     *
     * @param quantityName Nom de la quantit (exemple: "Mass").
     * @param symbol Symbole de l'unit (exemple: "kg").
     */
    private BaseUnit(final String quantityName, final String symbol, final PrefixSet prefix) {
        super(quantityName, symbol, prefix);
    }
    
    /**
     * Cre une nouvelle unit de base. Les units de base n'existent
     * normalement qu'en nombre restreint et servent de "briques" 
     * toutes les autres units. Avant de crer une nouvelle unit
     * avec cette mthode, vrifiez s'il ne s'agit pas en ralit
     * d'une unit drive.
     *
     * @param quantityName Le nom de la quantit associe  l'unit.
     *        Des exemples de noms seraient "mass", "electric current",
     *        "temperature", etc.
     *
     * @param symbol Le symbole des units (par exemple "kg" pour la masse).
     *        Ce symbole est obligatoire et ne doit pas tre <code>null</code>.
     *
     * @param prefix Liste des prfix pouvant tre utiliss avec le symbole
     *        <code>symbol</code>, ou <code>null</code> s'il n'y en a pas.
     *
     * @return Une unit de base associe  la quantit <code>quantityName</code>
     *         avec le symbole <code>symbol</code> et les prfix <code>prefix</code>.
     *
     * @see DerivedUnit#getInstance
     * @see ScaledUnit#getInstance
     * @see OffsetUnit#getInstance
     */
    public static BaseUnit getInstance(final String quantityName, final String symbol, final PrefixSet prefix) {
        return (BaseUnit) new BaseUnit(quantityName, symbol, prefix).intern();
    }
    
    /**
     * Retourne le symbole {@link #symbol} sans son prfix. Cette mthode retourne habituellement
     * {@link #symbol}, ce qui est correct pour la presque totalit des units <code>BaseUnit</code>.
     * Dans le systme SI, la seule exception notable (qui justifie  elle seule l'existence de cette
     * mthode) est le kilogramme (symbole "kg").
     */
    String getUnprefixedSymbol() {
        return symbol.equals("kg") ? "g" : super.getUnprefixedSymbol();
    }
    
    /**
     * Renvoie une unit identique  celle-ci, mais
     * avec un nouveau symbole et de nouveaux prfix.
     *
     * @param  symbol Nouveau symbole reprsentant cette unit. Si ce
     *         paramtre est nul, un symbole par dfaut sera cr.
     * @param  prefix Liste des prfix autoriss pour le symbole.
     * @return La mme unit, mais avec le nouveau symbole. Peut tre
     *         <code>this</code>, mais ne sera jamais <code>null</code>.
     */
    public Unit rename(final String symbol, final PrefixSet prefix) { // CAST
        return getInstance(quantityName, symbol, prefix);
    }
    
    /**
     * lve cette unit  une puissance entire.
     *
     * @param power La puissance  laquelle lever cette unit.
     * @return Les units rsultant de l'lvation des units
     *         <code>this</code>  la puissance <code>power</code>.
     *
     * @see #multiply
     * @see #divide
     * @see #scale
     * @see #shift
     */
    public Unit pow(final int power) {
        switch (power) {
            case 0:  return DerivedUnit.DIMENSIONLESS;
            case 1:  return this;
            default: return DerivedUnit.getInstance(new Factor[] {Factor.getFactor(this, power)});
        }
    }
    
    /**
     * Multiplie cette unit par une autre unit.
     *
     * @param  that L'unit par laquelle multiplier cette unit.
     * @return Le produit de <code>this</code> par <code>that</code>.
     * @throws UnitException Si l'unit <code>that</code> est de la
     *         classe {@link OffsetUnit} ou d'une autre classe invalide.
     *
     * @see #pow
     * @see #divide
     * @see #scale
     * @see #shift
     */
    public Unit multiply(final Unit that) throws UnitException {
        return that.inverseMultiply(this);
    }
    
    /**
     * Multiply a derived unit by a base unit
     * (<code>that</code>*<code>this</code>).
     */
    Unit inverseMultiply(final DerivedUnit that) {
        return that.multiply(new Factor[] {Factor.getFactor(this, +1)});
    }
    
    /**
     * Multiply a base unit by another base unit
     * (<code>that</code>*<code>this</code>).
     */
    Unit inverseMultiply(final BaseUnit that) throws UnitException {
        final Factor[] factors;
        if (equalsIgnoreSymbol(that)) {
            factors=new Factor[] {
                Factor.getFactor(that, 2)
            };
        } else {
            factors=new Factor[] {
                Factor.getFactor(that, 1),
                Factor.getFactor(this, 1)
            };
        }
        return DerivedUnit.getInstance(factors);
    }
    
    /**
     * Divise cette unit par une autre unit.
     *
     * @param that L'unit par laquelle diviser cette unit.
     * @return Le quotient de <code>this</code> par <code>that</code>.
     * @throws UnitException Si l'unit <code>that</code> est de la
     *         classe {@link OffsetUnit} ou d'une autre classe invalide.
     *
     * @see #pow
     * @see #multiply
     * @see #scale
     * @see #shift
     */
    public Unit divide(final Unit that) throws UnitException {
        return that.inverseDivide(this);
    }
    
    /**
     * Divide a derived unit by a base unit
     * (<code>that</code>/<code>this</code>).
     */
    Unit inverseDivide(final DerivedUnit that) {
        return that.multiply(new Factor[] {Factor.getFactor(this, -1)});
    }
    
    /**
     * Divide a base unit by another base unit
     * (<code>that</code>/<code>this</code>).
     */
    Unit inverseDivide(final BaseUnit that) throws UnitException {
        if (!equalsIgnoreSymbol(that)) {
            return DerivedUnit.getInstance(new Factor[] {
                Factor.getFactor(that, +1),
                Factor.getFactor(this, -1)
            });
        } else {
            return DerivedUnit.DIMENSIONLESS;
        }
    }
    
    /**
     * Indique si les units <code>this</code> et <code>that</code> sont compatibles.
     * Si elles le sont, alors les mthodes <code>convert</code> ne lanceront jamais
     * d'exception pour ces units.
     *
     * @param that Autre units avec laquelle on veut
     *        vrifier si ces units sont compatibles.
     * @return <code>true</code> Si l'on garantie que les mthodes
     *         <code>convert</code> ne lanceront pas d'exceptions.
     */
    public boolean canConvert(final Unit        that) {return that.canConvert(this);} // Do not move in superclass
           boolean canConvert(final DerivedUnit that) {return that.canConvert(this);} // Do not move in superclass
           boolean canConvert(final BaseUnit    that) {return equalsIgnoreSymbol(that);}
    
    /**
     * Effectue la conversion d'une mesure exprime selon d'autres units. Par
     * exemple <code>METRE.convert(1,&nbsp;FOOT)</code> retournera <code>0.3048</code>.
     *
     * @param value La valeur exprime selon les autres units (<code>fromUnit</code>).
     * @param fromUnit Les autres units.
     * @return La valeur convertie selon ces units (<code>this</code>).
     * @throws UnitException Si les units ne sont pas compatibles.
     */
    public double convert(final double value, final Unit        fromUnit) throws UnitException {return fromUnit.inverseConvert(value, this);} // Do not move in superclass
           double convert(final double value, final DerivedUnit fromUnit) throws UnitException {return fromUnit.inverseConvert(value, this);} // Do not move in superclass
           double convert(final double value, final BaseUnit    fromUnit) throws UnitException {if (!equalsIgnoreSymbol(fromUnit)) throw incompatibleUnitsException(fromUnit); return value;}
    
    /**
     * Effectue sur-place la conversion de mesures exprimes selon d'autres
     * units. Les valeurs converties remplaceront les anciennes valeurs.
     *
     * @param  values En entr, les valeurs exprimes selon les autres units
     *         (<code>fromUnit</code>). En sortie, les valeurs exprimes selon
     *         ces units (<code>this</code>).
     * @param  fromUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    public void convert(final double[] values, final Unit        fromUnit) throws UnitException {fromUnit.inverseConvert(values, this);} // Do not move in superclass
           void convert(final double[] values, final DerivedUnit fromUnit) throws UnitException {fromUnit.inverseConvert(values, this);} // Do not move in superclass
           void convert(final double[] values, final BaseUnit    fromUnit) throws UnitException {if (!equalsIgnoreSymbol(fromUnit)) throw incompatibleUnitsException(fromUnit);}
    
    /**
     * Effectue sur-place la conversion de mesures exprimes selon d'autres
     * units. Les valeurs converties remplaceront les anciennes valeurs.
     * Notez que d'importantes erreurs d'arrondissement peuvent survenir
     * si <code>fromUnit</code> est de la classe {@link OffsetUnit}.
     *
     * @param  values En entr, les valeurs exprimes selon les autres
     *         units (<code>fromUnit</code>). En sortie, les valeurs exprimes
     *         selon ces units (<code>this</code>).
     * @param  fromUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    public void convert(final float[] values, final Unit        fromUnit) throws UnitException {fromUnit.inverseConvert(values, this);} // Do not move in superclass
           void convert(final float[] values, final DerivedUnit fromUnit) throws UnitException {fromUnit.inverseConvert(values, this);} // Do not move in superclass
           void convert(final float[] values, final BaseUnit    fromUnit) throws UnitException {if (!equalsIgnoreSymbol(fromUnit)) throw incompatibleUnitsException(fromUnit);}
    
    /**
     * Retourne un objet qui saura convertir selon ces units les valeurs exprimes
     * selon d'autres units. Cette mthode est avantageuse si on prvoie fare
     * plusieurs conversions, car la transformation  utiliser est dtermine une
     * fois pour toute.
     *
     * @param  fromUnit Units  partir de lesquel faire les conversions.
     * @return Une transformation des units <code>fromUnit</code>
     *         vers les units <code>this</code>. Cette mthode ne
     *         retourne jamais <code>null</code>.
     * @throws UnitException Si les units ne sont pas compatibles.
     */
    public UnitTransform getTransform(final Unit        fromUnit) throws UnitException {return fromUnit.getInverseTransform(this);} // Do not move in superclass
           UnitTransform getTransform(final DerivedUnit fromUnit) throws UnitException {return fromUnit.getInverseTransform(this);} // Do not move in superclass
           UnitTransform getTransform(final BaseUnit    fromUnit) throws UnitException {return IdentityTransform.getInstance(fromUnit, this);}
    
    /**
     * Convertit une mesure vers d'autre units. Par exemple
     * <code>METRE.inverseConvert(1,&nbsp;FOOT)</code> retournera
     * <code>3.2808</code>. Cette mthode est l'inverse de la mthode
     * {@link #convert(double,Unit)}.
     *
     * @param value La valeur exprime selon ces units (<code>this</code>).
     * @param toUnit Les autres units.
     * @return La valeur convertie selon les autres units (<code>toUnit</code>).
     * @throws UnitException Si les units ne sont pas compatibles.
     */
    protected double inverseConvert(final double value, final Unit        toUnit) throws UnitException {return toUnit.convert(value, this);} // Do not move in superclass
              double inverseConvert(final double value, final DerivedUnit toUnit) throws UnitException {return toUnit.convert(value, this);} // Do not move in superclass
              double inverseConvert(final double value, final BaseUnit    toUnit) throws UnitException {if (!equalsIgnoreSymbol(toUnit)) throw toUnit.incompatibleUnitsException(this); return value;}
    
    /**
     * Effectue sur-place la conversion de mesures vers d'autres units.
     * Les valeurs converties remplaceront les anciennes valeurs. Cette
     * mthode est l'inverse de la mthode {@link #convert(double[],Unit)}.
     *
     * @param  values En entr, les valeur exprimes selon ces units
     *         (<code>this</code>). En sortie, les valeurs exprimes
     *         selon les autres units (<code>toUnit</code>).
     * @param  toUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    protected void inverseConvert(final double[] values, final Unit        toUnit) throws UnitException {toUnit.convert(values, this);} // Do not move in superclass
              void inverseConvert(final double[] values, final DerivedUnit toUnit) throws UnitException {toUnit.convert(values, this);} // Do not move in superclass
              void inverseConvert(final double[] values, final BaseUnit    toUnit) throws UnitException {if (!equalsIgnoreSymbol(toUnit)) throw toUnit.incompatibleUnitsException(this);}
    
    /**
     * Effectue sur-place la conversion de mesures vers d'autres units.
     * Les valeurs converties remplaceront les anciennes valeurs. Cette
     * mthode est l'inverse de la mthode {@link #convert(float[],Unit)}.
     *
     * @param  values En entr, les valeur exprimes selon ces units
     *         (<code>this</code>). En sortie, les valeurs exprimes
     *         selon les autres units (<code>toUnit</code>).
     * @param  toUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    protected void inverseConvert(final float[] values, final Unit        toUnit) throws UnitException {toUnit.convert(values, this);} // Do not move in superclass
              void inverseConvert(final float[] values, final DerivedUnit toUnit) throws UnitException {toUnit.convert(values, this);} // Do not move in superclass
              void inverseConvert(final float[] values, final BaseUnit    toUnit) throws UnitException {if (!equalsIgnoreSymbol(toUnit)) throw toUnit.incompatibleUnitsException(this);}
    
    /**
     * Retourne un objet qui saura convertir selon d'autres units les
     * valeurs exprimes selon ces units. Cette mthode est l'inverse
     * de {@link #getTransform}.
     *
     * @param  toUnit Units vers lesquel faire les conversions.
     * @return Une transformation des units <code>this</code>
     *         vers les units <code>toUnit</code>. Cette mthode
     *         ne retourne jamais <code>null</code>.
     * @throws UnitException Si les units ne sont pas compatibles.
     */
    protected UnitTransform getInverseTransform(final Unit        toUnit) throws UnitException {return toUnit.getTransform(this);} // Do not move in superclass
              UnitTransform getInverseTransform(final DerivedUnit toUnit) throws UnitException {return toUnit.getTransform(this);} // Do not move in superclass
              UnitTransform getInverseTransform(final BaseUnit    toUnit) throws UnitException {return IdentityTransform.getInstance(this, toUnit);}
    
    /**
     * Vrifie si cette unit est identique  une autre. Deux units sont considrs
     * identiques s'ils ont le mme champ {@link #quantityName}. Des symboles et prefix
     * diffrents sont autoriss.
     */
    final boolean equalsIgnoreSymbol(final BaseUnit unit) {
        return (unit==this) || quantityName.equals(unit.quantityName);
    }
    
    /**
     * Indique si deux units sont gales, en ignorant leurs symboles.
     * La comparaison ne prend en compte que les quantits telles que
     * retournes par {@link #getQuantityName}.
     */
    public boolean equalsIgnoreSymbol(final Unit unit) {
        return (unit instanceof BaseUnit) && equalsIgnoreSymbol((BaseUnit) unit);
    }
    
    /**
     * Retourne un code
     * pour cette unit.
     */
    public int hashCode() {
        return quantityName.hashCode();
    }
}
