/*
 * Units - Temporary implementation for Geotools 2
 * Copyright (C) 1998 University Corporation for Atmospheric Research (Unidata)
 *               1998 Bill Hibbard & al. (VisAD)
 *               1999 Pches et Ocans Canada
 *               2000 Institut de Recherche pour le Dveloppement
 *               2002 Centre for Computational Geography
 *
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *    Library General Public License for more details (http://www.gnu.org/).
 *
 *
 *    This package is inspired from the units package of VisAD.
 *    Unidata and Visad's work is fully acknowledged here.
 *
 *                   THIS IS A TEMPORARY CLASS
 *
 *    This is a placeholder for future <code>Unit</code> class.
 *    This skeleton will be removed when the real classes from
 *    JSR-108: Units specification will be publicly available.
 */
package org.geotools.units;

// Divers
import java.io.ObjectStreamException;
import org.geotools.resources.rsc.Resources;
import org.geotools.resources.rsc.ResourceKeys;


/**
 * Classe reprsentant des units dcales par rapport  d'autres units.
 * Les valeurs exprimes selon les units d'un objet <code>OffsetUnit</code>
 * peuvent tre converties vers le systme d'units {@link #unit}  l'aide
 * de l'quation suivante:
 *
 * <blockquote><pre>
 * <var>x</var><sub>{@link #unit}</sub>&nbsp;=&nbsp;{@link #offset}&nbsp;+&nbsp;<var>x</var><sub><code>this</code></sub>
 * </pre></blockquote>
 *
 * Les objets <code>OffsetUnit</code> permettent de faire des conversions entre
 * diffrentes units, par exemple des degrs Kelvin aux degrs Celsius. Cette
 * classe n'ayant pas de constructeur publique, la seule faon d'obtenir des
 * units dcales est d'utiliser les mthodes {@link #getInstance} ou {@link #offset}.
 *
 * <p><strong>NOTE: Des erreurs d'arrondissements importants peuvent survenir lors des
 * conversions d'units utilisant des objets <code>OffsetUnit</code></strong>. La densit Sigma-T,
 * utilise en ocanographie, en est un bon exemple. Cette densit Sigma-T est la densit de l'eau
 * de mer auquel on a retranche 1000&nbsp;kg/m. Par exemple une eau de mer de densit 1024&nbsp;kg/m
 * a une densit Sigma-T de 24&nbsp;kg/m. Dans un nombre rel de type <code>float</code>, un nombre
 * de l'ordre de 24 peut tre mmoris avec environ 6 chiffres aprs la virgules. Mais en revanche,
 * un nombre de l'ordre de 1024 ne peut tre mmoris qu'avec environ 3 ou 4 chiffres aprs la virgule.
 * Les conversions des densits Sigma-T en densit ordinaires (qui consistent  ajouter 1000 kg/m)
 * risquent donc de se traduire par une perte de 2 ou 3 chiffres significatifs de cette densit. Une faon
 * d'viter ce problme est de convertir tous les tableaux de <code>float</code> en <code>double</code>
 * avant de convertir les units.</p>
 *
 * @version 1.0
 * @author Steven R. Emmerson
 * @author Bill Hibbard
 * @author Martin Desruisseaux
 */
/*public*/
final class OffsetUnit extends Unit {
    /**
     * Serial number for interoperability with different versions.
     */
    private static final long serialVersionUID = 6259444767590765138L;
    
    /**
     * La constante  ajouter aux mesures exprimes selon ces
     * units pour obtenir des mesures exprimes selon les units
     * {@link #unit}. Cette constante sera par exemple 273.15 pour
     * passer des degrs Celsius ou degrs Kelvin.
     */
    public final double offset;
    
    /**
     * Les units vers lesquelles
     * se feront les conversions.
     */
    public final Unit unit;
    
    /**
     * Construit une unit dcale par rapport
     *  l'unit spcifie, avec le symbole
     * spcifi.
     *
     * @param  offset La constante du dcalage.
     * @param  unit L'unit associe. Pour des raisons
     *         d'efficacit, cette unit ne devrait pas
     *         tre un objet <code>OffsetUnit</code>.
     * @param  symbol Le symbole associ  cette unit.
     * @param  prefix Liste des prfix pouvant tre placs devant le symbole
     *         <code>symbol</code>, ou <code>null</code> s'il n'y en a pas.
     */
    private OffsetUnit(final double offset, final Unit unit, final String symbol, final PrefixSet prefix) {
        super(symbol, prefix);
        this.offset = offset;
        this.unit   = unit;
        
        if (unit==null) {
            throw new NullPointerException(Resources.format(ResourceKeys.ERROR_NO_UNIT));
        }
        if (Double.isNaN(offset) || Double.isInfinite(offset)) {
            throw new IllegalArgumentException(Resources.format(
                                               ResourceKeys.ERROR_NOT_A_NUMBER_$1,
                                               new Double(offset)));
        }
    }
    
    /**
     * Cre une nouvelle unit de mme dimension que l'unit spcifie, mais dont les mesures
     * seront dcales d'une constante. Les conversions d'units se feront par l'quation suivante:
     *
     * <center>
     *     <var>x</var><sub><code>unit</code></sub> =
     *     <var>x</var><sub><code>new</code></sub> + <code>offset</code>
     * </center>
     *
     * o <var>x</var><sub><code>unit</code></sub> reprsente une quantit mesure
     * selon les units <code>unit</code> et <var>x</var><sub><code>new</code></sub>
     * reprsente une quantit mesure selon les units retournes par cette mthode.
     *
     * Par exemple on pourrait construire des units de
     * temprature en degrs celsius avec le code suivant:
     *
     * <blockquote><pre>
     *     Unit CELSIUS=OffsetUnit.<strong>getInstance</strong>(273.15, KELVIN);
     * </blockquote></pre>
     *
     * Ce code signifie que 0C correspond  273.15K.
     *
     * @param  offset Constante qu'il faudra additionner aux mesures
     *         exprimes selon les nouvelles units pour les convertir
     *         dans les units <code>unit</code>.
     * @param  unit Units par rapport  lesquelles
     *         dcaler les nouvelles units.
     * @return Les units cres, ou <code>unit</code> si le
     *         paramtre <code>offset</code> est gal  0.
     *
     * @see BaseUnit#getInstance
     * @see DerivedUnit#getInstance
     * @see ScaledUnit#getInstance
     */
    public static Unit getInstance(final double offset, final Unit unit) {
        return getInstance(offset, unit, null, null);
    }
    
    /**
     * Cre une nouvelle unit de mme dimension que l'unit spcifie, mais dont les mesures
     * seront dcales d'une constante. Les conversions d'units se feront par l'quation suivante:
     *
     * <center>
     *     <var>x</var><sub><code>unit</code></sub> =
     *     <var>x</var><sub><code>new</code></sub> + <code>offset</code>
     * </center>
     *
     * o <var>x</var><sub><code>unit</code></sub> reprsente une quantit mesure
     * selon les units <code>unit</code> et <var>x</var><sub><code>new</code></sub>
     * reprsente une quantit mesure selon les units retournes par cette mthode.
     *
     * Par exemple on pourrait construire des units de
     * temprature en degrs celsius avec le code suivant:
     *
     * <blockquote><pre>
     *     Unit CELSIUS=OffsetUnit.<strong>getInstance</strong>(273.15, KELVIN, "C", null);
     * </blockquote></pre>
     *
     * Ce code signifie que 0C correspond  273.15K.
     *
     * @param  offset Constante qu'il faudra additionner aux mesures
     *         exprimes selon les nouvelles units pour les convertir
     *         dans les units <code>unit</code>.
     * @param  unit Units par rapport  lesquelles
     *         dcaler les nouvelles units.
     * @param  symbol Le symbole qui reprsentera les units cres. Si
     *         nul, alors un symbole par dfaut sera cr. Dans l'exemple
     *         prcdent, ce symbole par dfaut serait "+273.15&nbsp;K"
     * @param  prefix Liste des prfix pouvant tre placs devant le symbole
     *         <code>symbol</code>, ou <code>null</code> s'il n'y en a pas.
     * @return Les units cres, ou <code>unit</code> si le
     *         paramtre <code>offset</code> est gal  0.
     *
     * @see BaseUnit#getInstance
     * @see DerivedUnit#getInstance
     * @see ScaledUnit#getInstance
     */
    public static Unit getInstance(final double offset, final Unit unit, final String symbol, final PrefixSet prefix) {
        if (offset==0) {
            if (unit==null) {
                throw new NullPointerException(Resources.format(ResourceKeys.ERROR_NO_UNIT));
            }
            if (symbol!=null) {
                // TODO: Que faire si le symbole spcifi
                //       n'est pas le symbole de 'unit'?
            }
            return unit;
        }
        if (symbol==null) {
            return new OffsetUnit(offset, unit, UnitFormat.DEFAULT.formatOffset(offset, unit, new StringBuffer()).toString(), null).internIgnoreSymbol();
        } else {
            return new OffsetUnit(offset, unit, symbol, prefix).intern();
        }
    }
    
    /**
     * Renvoie une unit identique  celle-ci, mais
     * avec un nouveau symbole et de nouveaux prfix.
     *
     * @param  symbol Nouveau symbole reprsentant cette unit. Si ce
     *         paramtre est nul, un symbole par dfaut sera cr.
     * @param  prefix Liste des prfix autoriss pour le symbole.
     * @return La mme unit, mais avec le nouveau symbole. Peut tre
     *         <code>this</code>, mais ne sera jamais <code>null</code>.
     */
    public Unit rename(final String symbol, final PrefixSet prefix) { // CAST
        return getInstance(offset, unit, symbol, prefix);
    }
    
    /**
     * Retourne la quantit que reprsente cette unit. Les quantits sont des chanes de
     * caractres qui dcrivent le paramtre physique mesur, comme "mass" ou "speed". Si
     * aucune quantit n'est dfinie pour cette unit, retourne <code>null</code>.
     */
    public String getQuantityName() {
        return unit.getQuantityName();
    }
    
    /**
     * Cre une nouvelle unit proportionnelle  cette unit. Par exemple pour
     * convertir en millicelsius des tempratures exprimes en degrs Celsius,
     * il faut les multiplier par 1000. On peut exprimer cette relation par le
     * code <code>Unit&nbsp;mdegC=degC.scale(0.001)</code>.
     *
     * @param  amount Facteur par lequel il faudra diviser les valeurs
     *         exprimes selon ces units pour obtenir des valeurs
     *         exprimes selon les nouvelles units.
     * @return Les nouvelles units.
     *
     * @see #shift
     */
    public Unit scale(final double amount) {
        return (amount==1) ? this : getInstance(offset/amount, unit.scale(amount));
    }
    
    /**
     * Cre une nouvelle unit dcale par rapport  cette unit. Par exemple
     * pour convertir des degrs Kelvin en degrs Celsius, il faut soustraire
     * 273.15 aux degrs Kelvin. On peut exprimer cette relation par le code
     * <code>Unit&nbsp;celsius=kelvin.shift(273.15)</code>.
     *
     * @param  offset Constante  soustraire aux valeurs exprimes selon ces
     *         units pour obtenir des valeurs exprimes selon les nouvelles
     *         units.
     * @return Les nouvelles units.
     *
     * @see #scale
     */
    public Unit shift(final double offset) {
        return (offset==0) ? this : getInstance(this.offset+offset, unit);
    }
    
    /**
     * Indique si les units <code>this</code> et <code>that</code> sont compatibles.
     * Si elles le sont, alors les mthodes <code>convert</code> ne lanceront jamais
     * d'exception pour ces units.
     *
     * @param that Autre units avec laquelle on veut
     *        vrifier si ces units sont compatibles.
     * @return <code>true</code> Si l'on garantie que les mthodes
     *         <code>convert</code> ne lanceront pas d'exceptions.
     */
    public boolean canConvert(final Unit that) {
        return unit.canConvert(that);
    }
    
    /**
     * Effectue la conversion d'une mesure exprime selon d'autres units. Par
     * exemple <code>KELVIN.convert(15, CELSIUS)</code> retournera <code>288.15</code>.
     *
     * @param value    La valeur exprime selon les autres units (<code>fromUnit</code>).
     * @param fromUnit Les autres units.
     * @return         La valeur convertie selon ces units (<code>this</code>).
     * @throws         UnitException Si les units ne sont pas compatibles.
     */
    public double convert(final double value, final Unit fromUnit) throws UnitException {
        if (fromUnit==this) {
            return value; // sligh optimization
        }
        return unit.convert(value, fromUnit)-offset;
    }
    
    /**
     * Effectue sur-place la conversion de mesures exprimes selon d'autres
     * units. Les valeurs converties remplaceront les anciennes valeurs.
     *
     * @param  values En entr, les valeurs exprimes selon les autres units
     *         (<code>fromUnit</code>). En sortie, les valeurs exprimes selon ces
     *         units (<code>this</code>).
     * @param  fromUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    public void convert(final double[] values, final Unit fromUnit) throws UnitException {
        if (!equalsIgnoreSymbol(fromUnit)) {
            unit.convert(values, fromUnit);
            for (int i=0; i<values.length; i++) {
                values[i] -= offset;
            }
        }
    }
    
    /**
     * Effectue sur-place la conversion de mesures exprimes selon d'autres
     * units. Les valeurs converties remplaceront les anciennes valeurs.
     * Notez que d'importantes erreurs d'arrondissement peuvent survenir
     * si <code>fromUnit</code> est de la classe {@link OffsetUnit}.
     *
     * @param  values En entr, les valeurs exprimes selon les autres
     *         units (<code>fromUnit</code>). En sortie, les valeurs exprimes
     *         selon ces units (<code>this</code>).
     * @param  fromUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    public void convert(final float[] values, final Unit fromUnit) throws UnitException {
        if (!equalsIgnoreSymbol(fromUnit)) {
            unit.convert(values, fromUnit);
            for (int i=0; i<values.length; i++) {
                values[i] = (float) (values[i]-offset);
            }
        }
    }
    
    /**
     * Retourne un objet qui saura convertir selon ces units les valeurs exprimes
     * selon d'autres units. Cette mthode est avantageuse si on prvoie fare
     * plusieurs conversions, car la transformation  utiliser est dtermine une
     * fois pour toute.
     *
     * @param  fromUnit Units  partir de lesquel faire les conversions.
     * @return Une transformation des units <code>fromUnit</code>
     *         vers les units <code>this</code>. Cette mthode ne
     *         retourne jamais <code>null</code>.
     * @throws UnitException Si les units ne sont pas compatibles.
     */
    public UnitTransform getTransform(final Unit fromUnit) throws UnitException {
        if (!equalsIgnoreSymbol(fromUnit)) {
            final UnitTransform tr=unit.getTransform(fromUnit);
            if (tr instanceof OffsetTransform) {
                return OffsetTransform.getInstance(fromUnit, this, ((OffsetTransform) tr).offset+offset);
            } else {
                return CompoundTransform.getInstance(tr, OffsetTransform.getInstance(unit, this, offset));
            }
        } else {
            return IdentityTransform.getInstance(fromUnit, this);
        }
    }
    
    /**
     * Convertit une mesure vers d'autre units.
     *
     * @param value  La valeur exprime selon ces units (<code>this</code>).
     * @param toUnit Les autres units.
     * @return       La valeur convertie selon les autres units (<code>toUnit</code>).
     * @throws       UnitException Si les units ne sont pas compatibles.
     */
    protected double inverseConvert(final double value, final Unit toUnit) throws UnitException {
        if (toUnit==this) {
            return value; // sligh optimization
        }
        return unit.inverseConvert(value+offset, toUnit);
    }
    
    /**
     * Effectue sur-place la conversion de mesures vers d'autres units.
     * Les valeurs converties remplaceront les anciennes valeurs. Cette
     * mthode est l'inverse de la mthode {@link #convert(double[],Unit)}.
     *
     * @param  values En entr, les valeur exprimes selon ces units
     *         (<code>this</code>). En sortie, les valeurs exprimes
     *         selon les autres units (<code>toUnit</code>).
     * @param  toUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    protected void inverseConvert(final double[] values, final Unit toUnit) throws UnitException {
        if (!equalsIgnoreSymbol(toUnit)) {
            if (unit.canConvert(toUnit)) {
                for (int i=0; i<values.length; i++) {
                    values[i] += offset;
                }
                unit.inverseConvert(values, toUnit);
            } else {
                throw toUnit.incompatibleUnitsException(this);
            }
        }
    }
    
    /**
     * Effectue sur-place la conversion de mesures vers d'autres units.
     * Les valeurs converties remplaceront les anciennes valeurs. Cette
     * mthode est l'inverse de la mthode {@link #convert(float[],Unit)}.
     * @param  values En entr, les valeur exprimes selon ces units
     *         (<code>this</code>). En sortie, les valeurs exprimes
     *         selon les autres units (<code>toUnit</code>).
     * @param  toUnit Les autres units.
     * @throws UnitException Si les units ne sont pas compatibles. Dans ce
     *         cas, aucun lment de <code>values</code> n'aura t modifi.
     */
    protected void inverseConvert(final float[] values, final Unit toUnit) throws UnitException {
        if (!equalsIgnoreSymbol(toUnit)) {
            if (unit.canConvert(toUnit)) {
                for (int i=0; i<values.length; i++) {
                    values[i] = (float) (values[i]+offset);
                }
                unit.inverseConvert(values, toUnit);
            } else {
                throw toUnit.incompatibleUnitsException(this);
            }
        }
    }
    
    /**
     * Retourne un objet qui saura convertir selon d'autres units les
     * valeurs exprimes selon ces units. Cette mthode est l'inverse
     * de {@link #getTransform}.
     *
     * @param  toUnit Units vers lesquel faire les conversions.
     * @return Une transformation des units <code>this</code>
     *         vers les units <code>toUnit</code>. Cette mthode
     *         ne retourne jamais <code>null</code>.
     * @throws UnitException Si les units ne sont pas compatibles.
     */
    protected UnitTransform getInverseTransform(final Unit toUnit) throws UnitException {
        if (!equalsIgnoreSymbol(toUnit)) {
            final UnitTransform tr=unit.getInverseTransform(toUnit);
            if (tr instanceof OffsetTransform) {
                return ScaledTransform.getInstance(this, toUnit, ((OffsetTransform) tr).offset-offset);
            } else {
                return CompoundTransform.getInstance(OffsetTransform.getInstance(this, unit, -offset), tr);
            }
        } else {
            return IdentityTransform.getInstance(this, toUnit);
        }
    }
    
    /**
     * Indique si deux units sont gales, en ignorant leurs symboles. Le
     * champs {@link #symbol} de chacune des deux units ne sera pas pris
     * en compte.
     */
    public boolean equalsIgnoreSymbol(final Unit unit) {
        if (unit instanceof OffsetUnit) {
            final OffsetUnit that = (OffsetUnit) unit;
            return Double.doubleToLongBits(this.offset)==Double.doubleToLongBits(that.offset) &&
                   this.unit.equalsIgnoreSymbol(that.unit);
        } else {
            return false;
        }
    }
    
    /**
     * Retourne un code
     * pour cette unit.
     */
    public int hashCode() {
        final long code=Double.doubleToLongBits(offset);
        return (int) code ^ (int) (code >>> 32) ^ unit.hashCode();
    }
}
