/*
 * Units - Temporary implementation for Geotools 2
 * Copyright (C) 1998 University Corporation for Atmospheric Research (Unidata)
 *               1998 Bill Hibbard & al. (VisAD)
 *               1999 Pches et Ocans Canada
 *               2000 Institut de Recherche pour le Dveloppement
 *               2002 Centre for Computational Geography
 *
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *    Library General Public License for more details (http://www.gnu.org/).
 *
 *
 *    This package is inspired from the units package of VisAD.
 *    Unidata and Visad's work is fully acknowledged here.
 *
 *                   THIS IS A TEMPORARY CLASS
 *
 *    This is a placeholder for future <code>Unit</code> class.
 *    This skeleton will be removed when the real classes from
 *    JSR-108: Units specification will be publicly available.
 */
package org.geotools.units;

// Entrs/sorties
import java.io.Serializable;
import java.io.ObjectStreamException;

// Divers
import org.geotools.util.WeakHashSet;


/**
 * Reprsentation d'un prfix du systme mtrique. Un objet <code>Prefix</code>
 * peut par exemple reprsenter des "centi" (symbole "c") comme dans "centimtres"
 * (symbole "cm"). La description du paquet <code>org.geotools.units</code> donne
 * une liste des prfix standards du systme SI.
 *
 * <p><em>Note: this class has a natural ordering that is inconsistent with equals.</em>
 * La mthode {@link #compareTo} ne compare que le champ {@link #amount}, tandis
 * que la mthode {@link #equals} compare tous les champs ({@link #name},
 * {@link #symbol} et {@link #amount}).</p>
 *
 * @version 1.0
 * @author Martin Desruisseaux
 */
/*public*/ 
final class Prefix implements Comparable, Serializable {
    /**
     * Serial number for interoperability with different versions.
     */
    private static final long serialVersionUID = 3289659964721709283L;
    
    /**
     * Banque des objets qui ont t prcdemment crs et
     * enregistrs par un appel  la mthode {@link #intern}.
     */
    static final WeakHashSet pool=new WeakHashSet();
    
    /**
     * Nom neutre du prfix. Le systme SI dfinit plusieurs noms de prfix, parmi lesquels on trouve
     * "milli", "centi" et "kilo". Certaines units (notamment des units du type {@link ScaledUnit})
     * pourront combiner leurs noms avec un nom de prfix. Par exemple le prfix "centi" (symbole "c")
     * pourra tre combin avec les units "mtres" (symbole "m") pour former les "centimtres" (symbole
     * "cm"). La chane <code>name</code> peut tre vide, mais ne sera jamais nulle. Notez enfin que
     * <code>name</code> est "language-neutral". Pour obtenir un nom dans la langue de l'utilisateur,
     * utilisez la mthode {@link #getLocalizedName}.
     */
    public final String name;
    
    /**
     * Symbole du prfix. La plupart des symboles de prfix n'ont qu'une seule lettre. Il s'agit
     * la plupart du temps de la premire lettre de <code>name</code>, parfois en majuscule. Les
     * majuscules et minuscules sont significatifs et trs importants. Par exemple le symbole "m"
     * est pour "milli" tandis que le symbole "M" est pour "mega".
     */
    public final String symbol;
    
    /**
     * Quantit reprsent par ce prfix. Pour les prfix SI, cette quantit est toujours une puissance de 10.
     * Par exemple pour les "kilo" (symbole 'k'), la quantit <code>amount</code> est 1000. Cette quantit ne
     * sera jamais <code>NaN</code> ni infinie.
     */
    public final double amount;
    
    /**
     * Construit un prfix temporaire. Ce constructeur ne sert qu' effectuer
     * des recherches dans une liste de prfix par {@link PrefixSet}.
     */
    Prefix(final double amount) {
        this.name   = "";
        this.symbol = "";
        this.amount = amount;
    }
    
    /**
     * Construit un nouveau prfix.
     *
     * @param name    Nom du prfix (par exemple "centi" comme dans "centimtres").
     * @param symbol  Symbole du prfix (par exemple "c" pour "centimtres").
     * @param amount  Quantit reprsent par ce prfix (par exemple 0.01 pour "c").
     */
    private Prefix(final String name, final String symbol, final double amount) {
        this.name   = name.trim();
        this.symbol = symbol.trim();
        this.amount = amount;
        if (!(amount>0) || Double.isInfinite(amount)) {
            throw new IllegalArgumentException();
        }
    }
    
    /**
     * Construit un nouveau prfix.
     *
     * @param name    Nom du prfix (par exemple "centi" comme dans "centimtres").
     * @param symbol  Symbole du prfix (par exemple "c" pour "centimtres").
     * @param amount  Quantit reprsent par ce prfix (par exemple 0.01 pour "c").
     */
    public static Prefix getPrefix(final String name, final String symbol, final double amount) {
        return new Prefix(name, symbol, amount).intern();
    }
    
    /**
     * Retourne le nom du prfix dans la langue de l'utilisateur.
     * Par exemple le prfix "deci" est crit "dci" en franais.
     */
    public String getLocalizedName() {
        return org.geotools.resources.units.Prefix.localize(name);
    }
    
    /**
     * Retourne le symbole du prfix. Cette mthode retourne
     * systmatiquement le champ {@link #symbol}.
     */
    public String toString() {
        return symbol;
    }
    
    /**
     * Compare deux prfix. Cette mthode compare les quantits {@link #amount} de faon  permettre un classement
     * des prfix en ordre croissant de quantit. Contrairement  la mthode {@link #equals}, <code>compareTo</code>
     * ne compare pas les noms et symboles des prfix. Ainsi, deux prfix reprsentant la mme quantit mais avec
     * des symboles diffrents seront considrs gaux par <code>compareTo</code>.
     */
    public int compareTo(final Object object) {
        final Prefix that = (Prefix) object;
        if (this.amount > that.amount) return +1;
        if (this.amount < that.amount) return -1;
        return 0;
    }
    
    /**
     * Indique si ce prfix est identique  l'objet spcifi.
     * Cette mthode retourne <code>true</code> si <code>object</code> est aussi un
     * objet <code>Prefix</code> et si les deux prfix ont les mmes nom et symbole
     * et reprsentent la mme quantit {@link #amount}.
     */
    public boolean equals(final Object object) {
        if (object==this) {
            return true; // slight optimisation
        }
        if (object instanceof Prefix) {
            final Prefix prefix = (Prefix) object;
            return Double.doubleToLongBits(amount)==Double.doubleToLongBits(prefix.amount) &&
            symbol         .equals          (prefix.symbol) &&
            name           .equals          (prefix.name);
        }
        return false;
    }
    
    /**
     * Retourne un code reprsentant ce prfix.
     */
    public int hashCode() {
        final long code=Double.doubleToLongBits(amount);
        return (int) code ^ (int) (code >>> 32);
    }
    
    /**
     * Retourne un exemplaire unique de ce prfix. Une banque de prfix, initialement
     * vide, est maintenue de faon interne par la classe <code>Prefix</code>. Lorsque
     * la mthode <code>intern</code> est appele, elle recherchera un prfix gale 
     * <code>this</code> au sens de la mthode {@link #equals}. Si un tel prfix est
     * trouv, il sera retourn. Sinon, le prfix <code>this</code> sera ajout  la
     * banque de donnes en utilisant une rfrence faible et cette mthode retournera
     * <code>this</code>.
     * <br><br>
     * De cette mthode il s'ensuit que pour deux prfix <var>u</var> et <var>v</var>,
     * la condition <code>u.intern()==v.intern()</code> sera vrai si et seulement si
     * <code>u.equals(v)</code> est vrai.
     */
    private final Prefix intern() {
        return (Prefix) pool.canonicalize(this);
    }
    
    /**
     * Aprs la lecture d'une unit, vrifie si ce prfix
     * apparat dj dans la banque des prfix. Si oui,
     * l'exemplaire de la banque sera retourn plutt
     * que de garder inutilement le prfix courant comme copie.
     */
    final Object readResolve() throws ObjectStreamException {
        return intern();
    }
}
