/*
 * Units - Temporary implementation for Geotools 2
 * Copyright (C) 1998 University Corporation for Atmospheric Research (Unidata)
 *               1998 Bill Hibbard & al. (VisAD)
 *               1999 Pches et Ocans Canada
 *               2000 Institut de Recherche pour le Dveloppement
 *               2002 Centre for Computational Geography
 *
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *    Library General Public License for more details (http://www.gnu.org/).
 *
 *
 *    This package is inspired from the units package of VisAD.
 *    Unidata and Visad's work is fully acknowledged here.
 *
 *                   THIS IS A TEMPORARY CLASS
 *
 *    This is a placeholder for future <code>Unit</code> class.
 *    This skeleton will be removed when the real classes from
 *    JSR-108: Units specification will be publicly available.
 */
package org.geotools.units;

// Entrs/sorties
import java.io.Serializable;
import java.io.ObjectStreamException;

// Divers
import java.util.Arrays;
import org.geotools.util.WeakHashSet;


/**
 * Ensemble de prfix. Cette classe maintient une liste d'objets
 * {@link Prefix} en ordre croissant et sans doublons, c'est--dire qu'elle garanti
 * qu'il n'y aura pas deux prfix reprsentant la mme quantit {@link Prefix#amount}.
 *
 * @version 1.0
 * @author Martin Desruisseaux
 */
/*public*/ 
final class PrefixSet implements Serializable {
    /**
     * Serial number for interoperability with different versions.
     */
    private static final long serialVersionUID = -8301096197856692402L;
    
    /**
     * Banque des objets qui ont t prcdemment crs et
     * enregistrs par un appel  la mthode {@link #intern}.
     */
    private static final WeakHashSet pool=Prefix.pool;
    
    /**
     * Ensemble de prfix. Les prfix de cet ensemble doivent
     * obligatoirement tre un ordre croissant et sans doublons.
     */
    private final Prefix[] prefix;
    
    /**
     * Construit un ensemble de prfix. Le tableau <code>p</code>
     * sera copi, puis class. Les ventuels doublons seront limins.
     * Le tableau <code>p</code> original ne sera pas affect par ces
     * traitements.
     */
    private PrefixSet(final Prefix[] p) {
        final Prefix[] px=new Prefix[p.length];
        System.arraycopy(p, 0, px, 0, px.length);
        Arrays.sort(px);
        int length=px.length;
        for (int i=length; --i>=1;) {
            if (px[i].amount == px[i-1].amount) {
                px[i]=null;
                length--;
            }
        }
        int i=0;
        prefix=new Prefix[length];
        for (int j=0; j<px.length; j++) {
            if (px[j]!=null) {
                prefix[i++]=px[j];
            }
        }
        assert i==length;
    }
    
    /**
     * Construit un ensemble de prfix. Le tableau <code>p</code>
     * sera copi, puis class. Les ventuels doublons seront limins.
     * Le tableau <code>p</code> original ne sera pas affect par ces
     * traitements.
     */
    public static PrefixSet getPrefixSet(final Prefix[] p) {
        return new PrefixSet(p).intern();
    }
    
    /**
     * Retourne le prfix reprsent par le symbole spfifi.
     * Si aucun prfix ne correspond  ce symbole, retourne
     * <code>null</code>.
     *
     * @param  symbol Symbole du prfix recherch.
     * @return Prfix dsign par le symbole <code>symbol</code>.
     */
    public Prefix getPrefix(final String symbol) {
        for (int i=0; i<prefix.length; i++) {
            final Prefix p=prefix[i];
            if (symbol.equals(p.symbol)) {
                return p;
            }
        }
        return null;
    }
    
    /**
     * Retourne le prfix reprsentant une quantit gale ou infrieure  la quantit spcifie.
     * Si <code>amount</code> est infrieur  la plus petite quantit pouvant tre reprsent
     * par un prfix, alors cette mthode retourne <code>null</code>.
     */
    public Prefix getPrefix(double amount) {
        amount += 1E-8*Math.abs(amount); // Pour viter d'ventuelles erreurs d'arrondissements.
        int index=Arrays.binarySearch(prefix, new Prefix(amount));
        if (index<0) {
            index = ~index;
            if (index==0) {
                return null;
            }
            if (index>prefix.length) {
                index=prefix.length;
            }
            index--;
        }
        return prefix[index];
    }
    
    /**
     * Retourne une chane de caractres qui numre tous les prfix contenu dans
     * cet ensemble. La chane sera de la forme "milli(m),centi(c),dci(d),kilo(k)"
     * par exemple.
     */
    public String toString() {
        final StringBuffer buffer=new StringBuffer();
        for (int i=0; i<prefix.length; i++) {
            final Prefix p=prefix[i];
            final String name=p.getLocalizedName();
            final String symb=p.symbol;
            if (name.length()!=0 || symb.length()!=0) {
                if (buffer.length()!=0) {
                    buffer.append(',');
                }
                buffer.append(name);
                if (symb.length()!=0) {
                    buffer.append('(');
                    buffer.append(symb);
                    buffer.append(')');
                }
            }
        }
        return buffer.toString();
    }
    
    /**
     * Vrifie si cet ensemble est identique  l'objet <code>other</code>
     * spcifi. Deux ensembles sont considrs identiques s'ils contienent
     * les mmes prfix.
     */
    public boolean equals(final Object other) {
        if (other==this) return true; // slight optimisation
        if (other instanceof PrefixSet) {
            final Prefix[] array = ((PrefixSet) other).prefix;
            if (prefix.length == array.length) {
                for (int i=0; i<array.length; i++) {
                    if (!prefix[i].equals(array[i])) {
                        return false;
                    }
                }
                return true;
            }
        }
        return false;
    }
    
    /**
     * Retourne un code reprsentant cet ensemble de prfix.
     */
    public int hashCode() {
        int code=prefix.length << 1;
        for (int i=0; i<prefix.length; i+=5) {
            code += prefix[i].hashCode();
        }
        return code;
    }
    
    /**
     * Retourne un exemplaire unique de cet ensemble de prfix. Une banque de prfix, initialement
     * vide, est maintenue de faon interne par la classe <code>PrefixSet</code>. Lorsque la mthode
     * <code>intern</code> est appele, elle recherchera des prfix gaux  <code>this</code> au
     * sens de la mthode {@link #equals}. Si de tels prfix sont trouvs, ils seront retourns.
     * Sinon, les prfix <code>this</code> seront ajouts  la banque de donnes en utilisant une
     * {@link java.lang.ref.WeakReference rfrence faible} et cette mthode retournera <code>this</code>.
     * <br><br>
     * De cette mthode il s'ensuit que pour deux ensembles de prfix <var>u</var> et <var>v</var>,
     * la condition <code>u.intern()==v.intern()</code> sera vrai si et seulement si
     * <code>u.equals(v)</code> est vrai.
     */
    private final PrefixSet intern() {
        return (PrefixSet) pool.canonicalize(this);
    }
    
    /**
     * Aprs la lecture d'une unit, vrifie si ce prfix
     * apparat dj dans la banque des prfix {@link #PREFIX}.
     * Si oui, l'exemplaire de la banque sera retourn plutt
     * que de garder inutilement le prfix courant comme copie.
     */
    final Object readResolve() throws ObjectStreamException {
        return intern();
    }
}
