/*
 * Units - Temporary implementation for Geotools 2
 * Copyright (C) 1998 University Corporation for Atmospheric Research (Unidata)
 *               1998 Bill Hibbard & al. (VisAD)
 *               1999 Pches et Ocans Canada
 *               2000 Institut de Recherche pour le Dveloppement
 *               2002 Centre for Computational Geography
 *
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *    Library General Public License for more details (http://www.gnu.org/).
 *
 *
 *    This package is inspired from the units package of VisAD.
 *    Unidata and Visad's work is fully acknowledged here.
 *
 *                   THIS IS A TEMPORARY CLASS
 *
 *    This is a placeholder for future <code>Unit</code> class.
 *    This skeleton will be removed when the real classes from
 *    JSR-108: Units specification will be publicly available.
 */
package org.geotools.units;

// Entrs/sorties
import java.io.Serializable;
import java.io.ObjectStreamException;

// Divers
import org.geotools.util.WeakHashSet;
import org.geotools.resources.Utilities;


/**
 * Reprsente une transformation entre deux units. Par convention, tout
 * les objets <code>UnitTransform</code> sont toujours imutables. Il est
 * donc scuritaire de partager plusieurs rfrences vers le mme objet.
 *
 * @version 1.0
 * @author Martin Desruisseaux
 */
public abstract class UnitTransform implements Serializable {
    /**
     * Serial number for interoperability with different versions.
     */
    private static final long serialVersionUID = 59496814325077015L;

    /**
     * Banque des objets qui ont t prcdemment crs et
     * enregistrs par un appel  la mthode {@link #intern}.
     */
    private static final WeakHashSet pool=Prefix.pool;

    /**
     * Unit selon laquelle seront
     * exprimes les valeurs initiales.
     */
    public final Unit fromUnit;

    /**
     * Unit selon laquelle seront
     * exprimes les valeurs finales.
     */
    public final Unit toUnit;

    /**
     * Construit un objet qui aura la charge de convertir
     * des donnes exprimes selon les units spcifies.
     */
    /*protected*/ UnitTransform(final Unit fromUnit, final Unit toUnit) {
        this.fromUnit = fromUnit;
        this.toUnit   = toUnit;
    }

    /**
     * Indique si cette transformation affine reprsente une transformation idntite.
     * L'implmentation par dfaut retourne <code>toUnit.equalsIgnoreSymbol(fromUnit)</code>.
     */
    public boolean isIdentity() {
        return toUnit.equalsIgnoreSymbol(fromUnit);
    }

    /**
     * Effectue la conversion d'units d'une valeur.
     * @param value Valeur exprime selon les units {@link #fromUnit}.
     * @return Valeur exprime selon les units {@link #toUnit}.
     */
    public abstract double convert(double value);

    /**
     * Effectue la conversion d'units d'un tableaux de valeurs.
     * L'implmentation par dfaut appelle {@link #convert(double)}
     * dans une boucle. Les classes drives devraient redfinir cette
     * mthode avec une implmentation plus efficace.
     *
     * @param values Valeurs exprimes selon les units {@link #fromUnit}.
     *        Elles seront converties sur place en valeurs exprimes selon
     *        les units {@link #toUnit}.
     */
    public void convert(final double[] values) {
        for (int i=0; i<values.length; i++) {
            values[i]=convert(values[i]);
        }
    }

    /**
     * Effectue la conversion d'units d'un tableaux de valeurs.
     * L'implmentation par dfaut appelle {@link #convert(double)}
     * dans une boucle. Les classes drives devraient redfinir cette
     * mthode avec une implmentation plus efficace.
     *
     * @param values Valeurs exprimes selon les units {@link #fromUnit}.
     *        Elles seront converties sur place en valeurs exprimes selon
     *        les units {@link #toUnit}.
     */
    public void convert(final float[] values) {
        for (int i=0; i<values.length; i++) {
            values[i]=(float) convert(values[i]);
        }
    }

    /**
     * Effectue la conversion inverse d'units d'une valeur.
     * @param value Valeur exprime selon les units {@link #toUnit}.
     * @return Valeur exprime selon les units {@link #fromUnit}.
     */
    public abstract double inverseConvert(double value);

    /**
     * Effectue la conversion inverse d'units d'un tableaux de valeurs.
     * L'implmentation par dfaut appelle {@link #inverseConvert(double)}
     * dans une boucle. Les classes drives devraient redfinir cette
     * mthode avec une implmentation plus efficace.
     *
     * @param values Valeurs exprimes selon les units {@link #toUnit}.
     *        Elles seront converties sur place en valeurs exprimes selon
     *        les units {@link #fromUnit}.
     */
    public void inverseConvert(final double[] values) {
        for (int i=0; i<values.length; i++) {
            values[i]=inverseConvert(values[i]);
        }
    }

    /**
     * Effectue la conversion inverse d'units d'un tableaux de valeurs.
     * L'implmentation par dfaut appelle {@link #inverseConvert(double)}
     * dans une boucle. Les classes drives devraient redfinir cette
     * mthode avec une implmentation plus efficace.
     *
     * @param values Valeurs exprimes selon les units {@link #toUnit}.
     *        Elles seront converties sur place en valeurs exprimes selon
     *        les units {@link #fromUnit}.
     */
    public void inverseConvert(final float[] values) {
        for (int i=0; i<values.length; i++) {
            values[i]=(float) inverseConvert(values[i]);
        }
    }

    /**
     * Retourne un exemplaire unique de cette transformation. Une banque de
     * transformation, initialement vide, est maintenue de faon interne par
     * la classe <code>UnitTransform</code>. Lorsque la mthode <code>intern</code>
     * est appelle, elle recherchera une transformation gale  <code>this</code>
     * au sens de la mthode {@link #equals}. Si une telle transformation fut trouve,
     * elle sera retourne. Sinon, la trsnsformation <code>this</code> sera ajoute 
     * la banque de donnes en utilisant une rfrence faible et cette mthode retournera
     * <code>this</code>.
     *
     * <p>De cette mthode il s'ensuit que pour deux transformations <var>u</var> et <var>v</var>,
     * la condition <code>u.intern()==v.intern()</code> sera vrai si et seulement si
     * <code>u.equals(v)</code> est vrai.</p>
     */
    /*public*/ final UnitTransform intern() {
        return (UnitTransform) pool.canonicalize(this);
    }

    /**
     * Indique si cet objet est identique  l'objet spcifi.
     * Les deux objets seront considirs identiques s'ils
     * sont de la mme classe et font les conversions  partir
     * de et vers les mmes units.
     */
    public boolean equals(final Object o) {
        if (o!=null && getClass().equals(o.getClass())) {
            final UnitTransform ut=(UnitTransform) o;
            return fromUnit.equals(ut.fromUnit) && toUnit.equals(ut.toUnit);
        }
        return false;
    }

    /**
     * Retourne un code reprsentant
     * cette transformation d'units.
     */
    public int hashCode() {
        return fromUnit.hashCode() ^ toUnit.hashCode();
    }

    /**
     * Retourne une chane de caractres reprsentant
     * cette transformation. La chane sera de la forme
     *
     * <code>UnitTransform[km/h&nbsp;-->&nbsp;m/s]</code>
     */
    public String toString() {
        return Utilities.getShortClassName(this)+'['+fromUnit+" --> "+toUnit+']';
    }
    
    /**
     * Aprs la lecture d'une transformation, vrifie si cette transformation apparat
     * dj dans la banque des units <code>pool</code>. Si oui, l'exemplaire de la banque
     * sera retourn plutt que de garder inutilement la transformation courante comme
     * copie.
     */
    final Object readResolve() throws ObjectStreamException {
        return intern();
    }
}
