/*
 * Geotools - OpenSource mapping toolkit (C) 2002, Centre for Computational
 * Geography
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 */

package org.geotools.data.shapefile.shp;

import java.io.*;
import java.nio.*;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.FileChannel;
import java.util.*;
import org.geotools.io.NIOBufferUtils;

/**
 * The general use of this class is: <CODE><PRE>
 * 
 * FileChannel in = new FileInputStream("thefile.dbf").getChannel();
 * ShapefileReader r = new ShapefileReader( in ) while (r.hasNext()) { Geometry
 * shape = (Geometry) r.nextRecord().shape() // do stuff } r.close();
 * 
 * </PRE></CODE> You don't have to immediately ask for the shape from the record. The
 * record will contain the bounds of the shape and will only read the shape
 * when the shape() method is called. This ShapefileReader.Record is the same
 * object every time, so if you need data from the Record, be sure to copy it.
 * 
 * @author jamesm
 * @author aaime
 * @author Ian Schneider
 */
public class ShapefileReader {
  
  /** The reader returns only one Record instance in its lifetime. The
   * record contains the current record information.
   */  
  public final class Record {
    int length;
    int number = 0;
    int offset;
    int start = 0;
    /** The minimum X value. */    
    public double minX;
    /** The minimum Y value. */    
    public double minY;
    /** The maximum X value. */    
    public double maxX;
    /** The maximum Y value. */    
    public double maxY;
    ShapeType type;
    int end = 0;
    Object shape = null;
    /** Fetch the shape stored in this record. */    
    public Object shape() {
      if (shape == null) {
        buffer.position(start);
        buffer.order(ByteOrder.LITTLE_ENDIAN);
        shape = handler.read(buffer,type);
      }
      return shape;
    }
    public int offset() {
      return offset;
    }
    /** A summary of the record. */    
    public String toString() {
			return "Record " + number
                + " length " + length
				+ " bounds " + minX + "," + minY
				+ " " + maxX + "," + maxY;
		}
  }
  
  private ShapeHandler handler;
  private ShapefileHeader header;
  private ReadableByteChannel channel;
  private ByteBuffer buffer;
  private ShapeType fileShapeType = ShapeType.UNDEFINED;
  private ByteBuffer headerTransfer;
  private final Record record = new Record();
  private final boolean randomAccessEnabled;
  
  /** Creates a new instance of ShapeFile.
   * @param channel The ReadableByteChannel this reader will use.
   * @param strict True to make the header parsing throw Exceptions if the version or magic number
   * are incorrect.
   * @throws IOException If problems arise.
   * @throws ShapefileException If for some reason the file contains invalid records.
   */
  public ShapefileReader(ReadableByteChannel channel, boolean strict) throws IOException, ShapefileException {
    this.channel = channel;
    randomAccessEnabled = channel instanceof FileChannel;
    init(strict);
  }
  
  /** Default constructor. Calls ShapefileReader(channel,true).
   * @param channel
   * @throws IOException
   * @throws ShapefileException
   */  
  public ShapefileReader(ReadableByteChannel channel) throws IOException, ShapefileException {
    this(channel,true);
  }
  
  // convenience to peak at a header
  /** A short cut for reading the header from the given channel.
   * @param channel The channel to read from.
   * @param strict True to make the header parsing throw Exceptions if the version or magic number
   * are incorrect.
   * @throws IOException If problems arise.
   * @return A ShapefileHeader object.
   */  
  public static ShapefileHeader readHeader(ReadableByteChannel channel,boolean strict) throws IOException {
    ByteBuffer buffer = ByteBuffer.allocateDirect(100);
    if ( fill(buffer, channel) == -1 ) {
      throw new EOFException("Premature end of header");
    }
    buffer.flip();
    ShapefileHeader header = new ShapefileHeader();
    header.read(buffer, strict);
    return header;
  }
  
  // ensure the capacity of the buffer is of size by doubling the original
  // capacity until it is big enough
  // this may be naiive and result in out of MemoryError as implemented...
  private ByteBuffer ensureCapacity(ByteBuffer buffer,int size) {
    // This sucks if you accidentally pass is a MemoryMappedBuffer of size 80M
    // like I did while messing around, within moments I had 1 gig of swap...
    if (buffer.isReadOnly()) {
      return buffer;
    }
    
    int limit = buffer.limit();
    while (limit < size) {
      limit *= 2;
    }
    if (limit != buffer.limit()) {
      //if (record.ready) {
        buffer = ByteBuffer.allocateDirect(limit);
      //}
      //else {
      //  throw new IllegalArgumentException("next before hasNext");
      //}
    }
    return buffer;
  }
  
  // for filling a ReadableByteChannel
  private static int fill(ByteBuffer buffer,ReadableByteChannel channel) throws IOException {
    int r = buffer.remaining();
    // channel reads return -1 when EOF or other error
    // because they a non-blocking reads, 0 is a valid return value!!
    while (buffer.remaining() > 0 && r != -1) {
      r = channel.read(buffer);
    }
    if (r == -1) {
      buffer.limit(buffer.position());
    }
    return r;
  }
  
  private void init(boolean strict) throws IOException,ShapefileException {
    header = readHeader(channel,strict);
    fileShapeType = header.getShapeType();
    handler = fileShapeType.getShapeHandler();
    
//    recordHeader = ByteBuffer.allocateDirect(8);
//    recordHeader.order(ByteOrder.BIG_ENDIAN);
    
    if (handler == null) {
      throw new IOException("Unsuported shape type:" + fileShapeType);
    }
    
    if (channel instanceof FileChannel) {
      FileChannel fc = (FileChannel) channel;
      buffer = fc.map(FileChannel.MapMode.READ_ONLY,0,fc.size());
      buffer.position(100);
    } else {
      // start with 8K buffer
      buffer = ByteBuffer.allocateDirect(8 * 1024);
      fill(buffer,channel);
      buffer.flip();
    }
    
    headerTransfer = ByteBuffer.allocate(8);
    headerTransfer.order(ByteOrder.BIG_ENDIAN);
    
    // make sure the record end is set now...
    record.end = buffer.position();
  }
  
  /** Get the header. Its parsed in the constructor.
   * @return The header that is associated with this file.
   */  
  public ShapefileHeader getHeader() {
    return header;
  }
  
  // do important cleanup stuff.
  // Closes channel !
  /** Clean up any resources. Closes the channel.
   * @throws IOException If errors occur while closing the channel.
   */  
  public void close() throws IOException {
    if (channel.isOpen()) {
      channel.close();
    }
    if(buffer instanceof MappedByteBuffer) {
        NIOBufferUtils.clean(buffer);
    }
    channel = null;
    header = null;
  }
  
  public boolean supportsRandomAccess() {
    return randomAccessEnabled; 
  }
  
  /** If there exists another record. Currently checks the stream for the presence of
   * 8 more bytes, the length of a record. If this is true and the record indicates
   * the next logical record number, there exists more records.
   * @throws IOException
   * @return True if has next record, false otherwise.
   */  
  public boolean hasNext() throws IOException {
      
    // mark current position
    int position = buffer.position();
      
    // ensure the proper position, regardless of read or handler behavior
    buffer.position(record.end);
    
    // no more data left
    if (buffer.remaining() < 8) return false;
    
    // record headers in big endian
    buffer.order(ByteOrder.BIG_ENDIAN);

    // looks good
    boolean hasNext = buffer.getInt() == record.number + 1;
    
    // reset things to as they were
    buffer.position(position);
    
    return hasNext;
  }
  
  /**
   * Transfer (by bytes) the data at the current record to the ShapefileWriter.
   * @param bounds double array of length four for transfering the bounds into
   * @return The length of the record transfered in bytes
   */ 
  public int transferTo(ShapefileWriter writer,int recordNum,double[] bounds) throws IOException {

    buffer.position(record.end);
    buffer.order(ByteOrder.BIG_ENDIAN);
    
    buffer.getInt(); // record number
    int rl = buffer.getInt();
    int mark = buffer.position();
    int len = rl * 2;
    
    buffer.order(ByteOrder.LITTLE_ENDIAN);
    ShapeType recordType = ShapeType.forID(buffer.getInt());
    
    if (recordType.isMultiPoint()) {
      for (int i = 0; i < 4; i++) {
        bounds[i] = buffer.getDouble();
      }
    } 
    else if (recordType != recordType.NULL) {
      bounds[0] = bounds[1] = buffer.getDouble();
      bounds[2] = bounds[3] = buffer.getDouble();
    }

    // write header to shp and shx
    headerTransfer.position(0);
    headerTransfer.putInt(recordNum).putInt(rl).position(0);
    writer.shpChannel.write(headerTransfer);
    headerTransfer.putInt(0,writer.offset).position(0);
    writer.offset += rl + 4;
    writer.shxChannel.write(headerTransfer);
    
    // reset to mark and limit at end of record, then write
    buffer.position(mark).limit(mark + len);
    writer.shpChannel.write(buffer);
    buffer.limit(buffer.capacity());
    
    record.end = buffer.position();
    record.number++;
    
    return len;
  }
  
  /** Fetch the next record information.
   * @throws IOException
   * @return The record instance associated with this reader.
   */  
  public Record nextRecord() throws IOException {
      
    // need to update position
    buffer.position(record.end);
    
    // record header is big endian
    buffer.order(ByteOrder.BIG_ENDIAN);
    
    // read shape record header
    int recordNumber = buffer.getInt();
    // silly ESRI say contentLength is in 2-byte words
    // and ByteByffer uses bytes.
    // track the record location
    int recordLength = buffer.getInt() * 2;

    if (! buffer.isReadOnly() ) {
        // capacity is less than required for the record
        // copy the old into the newly allocated
        if (buffer.capacity() < recordLength + 8) {
          ByteBuffer old = buffer;
          // ensure enough capacity for one more record header
          buffer = ensureCapacity(buffer, recordLength + 8);
          buffer.put(old);
          fill(buffer,channel);
          buffer.position(0);
        } else 
        // remaining is less than record length
        // compact the remaining data and read again,
        // allowing enough room for one more record header
        if (buffer.remaining() < recordLength + 8) {
          buffer.compact();
          fill(buffer,channel);
          buffer.position(0);
        }
    }

    // shape record is all little endian
    buffer.order(ByteOrder.LITTLE_ENDIAN);
    
    // read the type, handlers don't need it
    ShapeType recordType = ShapeType.forID(buffer.getInt());
    
    // this usually happens if the handler logic is bunk,
    // but bad files could exist as well...
    if (recordType != ShapeType.NULL && recordType != fileShapeType) {
      throw new IllegalStateException("ShapeType changed illegally from " + fileShapeType + " to " + recordType);
    }

    // peek at bounds, then reset for handler
    // many handler's may ignore bounds reading, but we don't want to
    // second guess them...
    buffer.mark();
    if (recordType.isMultiPoint()) {
      record.minX = buffer.getDouble();
      record.minY = buffer.getDouble();
      record.maxX = buffer.getDouble();
      record.maxY = buffer.getDouble();
    } 
    else if (recordType != recordType.NULL) {
      record.minX = record.maxX = buffer.getDouble();
      record.minY = record.minY = buffer.getDouble();
    }
    buffer.reset();

    record.offset += record.length + (record.length == 0 ? 0 : 8);
    // update all the record info.
    record.length = recordLength;
    record.type = recordType;
    record.number = recordNumber;
    // remember, we read one int already...
    record.end = buffer.position() + recordLength - 4;
    // mark this position for the reader
    record.start = buffer.position();
    // clear any cached shape
    record.shape = null;

    return record;
  }
  
  public Object shapeAt(int offset) throws IOException,UnsupportedOperationException {
    if (randomAccessEnabled) {
      buffer.position(offset);
      hasNext();
      return nextRecord().shape();
    }
    throw new UnsupportedOperationException("Random Access not enabled");
  }
   
  public static void main(String[] args) throws Exception {
    FileChannel channel = new FileInputStream(args[0]).getChannel();
    ShapefileReader reader = new ShapefileReader(channel);
    System.out.println(reader.getHeader());
    while (reader.hasNext()) {
      System.out.println(reader.nextRecord().shape());
    }
  }
  
}
